# -*- coding: utf-8 -*-
# wasp_general/network/transport.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
import socket
import struct

from wasp_general.verify import verify_type, verify_value
from wasp_general.config import WConfig

from wasp_general.network.primitives import WIPV4SocketInfo, WIPV4Address, WNetworkIPV4


class WNetworkNativeTransportProto(metaclass=ABCMeta):
	""" This is interface for classes, that implement transport logic for network communication. "Native" means
	that these classes use socket objects directly.
	"""

	@abstractmethod
	@verify_type(config=WConfig)
	def server_socket(self, config):
		""" Return server socket. This socket is used for receiving requests and sending results. It is
		important, that the result can be polled by a IOLoop instance.

		:param config: server configuration
		:return: socket.socket
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(config=WConfig, close_fd=bool)
	def close_server_socket(self, config, close_fd=True):
		""" Close previously opened server socket. If no socket is opened - do nothing.

		:param config: server configuration
		:param close_fd: should this function close socket fd, or will it close by an external function?. It \
		is safer to pass True here.
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(config=WConfig)
	def client_socket(self, config):
		""" Return client socket. This socket is used for sending request and receiving results. It is
		important, that the result can be polled by a IOLoop instance.

		:param config: client configuration
		:return: socket.socket
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(config=WConfig, close_fd=bool)
	def close_client_socket(self, config, close_fd=True):
		""" Close previously opened client socket. If no socket is opened - do nothing.

		:param config: client configuration
		:param close_fd: should this function close socket fd, or will it close by an external function?. It \
		is safer to pass True here.
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(config=WConfig)
	def target_socket(self, config):
		""" Return socket information with server address. Mostly used for address validation.

		:param config: client configuration
		:return: WIPV4SocketInfo
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(config=WConfig)
	def bind_socket(self, config):
		""" Return socket information with address that server binds to.

		:param config: server configuration
		:return: WIPV4SocketInfo
		"""
		raise NotImplementedError('This method is abstract')


class WNetworkNativeTransportSocketConfig:
	""" Represent socket configuration settings.
	"""

	@verify_type(section=str, address_option=str, port_option=str)
	@verify_value(section=lambda x: len(x) > 0, address_option=lambda x: len(x) > 0, port_option=lambda x: len(x) > 0)
	def __init__(self, section, address_option, port_option):
		""" Construct new configuration settings

		:param section: section name
		:param address_option: address option name
		:param port_option: port option name
		"""
		self.section = section
		self.address_option = address_option
		self.port_option = port_option


class WNetworkNativeTransport(WNetworkNativeTransportProto, metaclass=ABCMeta):
	""" Basic WNetworkNativeTransportProto implementation. This class isn't ready to use, but it has general
	implementation for the most WNetworkNativeTransportProto methods.
	"""

	@verify_type(target_socket_config=WNetworkNativeTransportSocketConfig)
	@verify_type(bind_socket_config=WNetworkNativeTransportSocketConfig)
	def __init__(self, target_socket_config, bind_socket_config):
		""" Create new transport

		:param target_socket_config: configuration for client socket
		:param bind_socket_config: configuration for server socket
		"""

		self.__target_socket_config = target_socket_config
		self.__bind_socket_config = bind_socket_config

		self.__server_socket = None
		self.__client_socket = None

	@verify_type(config=WConfig)
	def target_socket(self, config):
		""" :meth:`.WNetworkNativeTransportProto.server_socket` method implementation
		"""
		address = config[self.__target_socket_config.section][self.__target_socket_config.address_option]
		port = config.getint(self.__target_socket_config.section, self.__target_socket_config.port_option)

		target = WIPV4SocketInfo(address, port)
		if target.address() is None or target.port() is None:
			raise ValueError('Invalid target address or port')
		return target

	@verify_type(config=WConfig)
	def bind_socket(self, config):
		""" :meth:`.WNetworkNativeTransportProto.bind_socket` method implementation
		"""
		address = config[self.__bind_socket_config.section][self.__bind_socket_config.address_option]
		port = config.getint(self.__bind_socket_config.section, self.__bind_socket_config.port_option)
		return WIPV4SocketInfo(address, port)

	@abstractmethod
	def _create_socket(self):
		""" Create general socket object, that can be used for client and/or server usage

		:return: socket.socket
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(config=WConfig)
	def create_server_socket(self, config):
		""" Create socket for server. (By default, same as WNetworkNativeTransport._create_socket)

		:param config: server configuration
		:return: socket.socket
		"""
		return self._create_socket()

	@verify_type(config=WConfig)
	def create_client_socket(self, config):
		""" Create socket for client. (By default, same as WNetworkNativeTransport._create_socket)

		:param config: client configuration
		:return: socket.socket
		"""
		return self._create_socket()

	@verify_type(config=WConfig)
	def server_socket(self, config):
		""" :meth:`.WNetworkNativeTransportProto.server_socket` method implementation
		"""
		if self.__server_socket is None:
			self.__server_socket = self.create_server_socket(config)
			self.__server_socket.bind(self.bind_socket(config).pair())
		return self.__server_socket

	@verify_type(config=WConfig, close_fd=bool)
	def close_server_socket(self, config, close_fd=True):
		""" :meth:`.WNetworkNativeTransportProto.close_server_socket` method implementation
		"""
		if close_fd is True:
			self.__server_socket.close()
		self.__server_socket = None

	@verify_type(config=WConfig)
	def client_socket(self, config):
		""" :meth:`.WNetworkNativeTransportProto.client_socket` method implementation
		"""
		if self.__client_socket is None:
			self.__client_socket = self.create_client_socket(config)
		return self.__client_socket

	@verify_type(config=WConfig, close_fd=bool)
	def close_client_socket(self, config, close_fd=True):
		""" :meth:`.WNetworkNativeTransportProto.close_client_socket` method implementation
		"""
		if close_fd is True:
			self.__client_socket.close()
		self.__client_socket = None


class WUDPNetworkNativeTransport(WNetworkNativeTransport, metaclass=ABCMeta):
	""" Basic UDP transport implementation
	"""

	def _create_socket(self):
		""" Create general UDP-socket

		:return: socket.socket
		"""
		return socket.socket(socket.AF_INET, socket.SOCK_DGRAM)


class WTCPNetworkNativeTransport(WNetworkNativeTransport):
	""" Basic TCP transport implementation
	"""

	def _create_socket(self):
		""" Create general TCP-socket

		:return: socket.socket
		"""
		return socket.socket(socket.AF_INET, socket.SOCK_STREAM)


class WBroadcastNetworkTransport(WUDPNetworkNativeTransport):
	""" Network transport, that uses IPv4 broadcast (UDP) communication
	"""

	@verify_type('paranoid', target_socket_config=WNetworkNativeTransportSocketConfig)
	@verify_type('paranoid', bind_socket_config=WNetworkNativeTransportSocketConfig)
	def __init__(self, target_socket_config, bind_socket_config):
		""" Create new broadcast transport
		"""
		WNetworkNativeTransport.__init__(self, target_socket_config, bind_socket_config)

	@verify_type('paranoid', config=WConfig)
	def target_socket(self, config):
		""" This method overrides :meth:`.WNetworkNativeTransport.target_socket` method. Do the same thing as
		basic method do, but also checks that the result address is IPv4 address.

		:param config: beacon configuration
		:return: WIPV4SocketInfo
		"""
		target = WNetworkNativeTransport.target_socket(self, config)
		if isinstance(target.address(), WIPV4Address) is False:
			raise ValueError('Invalid address for broadcast transport')
		return target

	@verify_type('paranoid', config=WConfig)
	def create_client_socket(self, config):
		""" Create client broadcast socket

		:param config: client configuration
		:return: socket.socket
		"""
		client_socket = WUDPNetworkNativeTransport.create_client_socket(self, config)
		client_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
		return client_socket


class WMulticastNetworkTransport(WUDPNetworkNativeTransport):
	""" Network transport, that uses IPv4 multicast communication
	"""

	@verify_type('paranoid', target_socket_config=WNetworkNativeTransportSocketConfig)
	@verify_type('paranoid', bind_socket_config=WNetworkNativeTransportSocketConfig)
	def __init__(self, target_socket_config, bind_socket_config):
		""" Create new multicast transport
		"""
		WUDPNetworkNativeTransport.__init__(self, target_socket_config, bind_socket_config)

	@verify_type('paranoid', config=WConfig)
	def target_socket(self, config):
		""" This method overrides :meth:`.WNetworkNativeTransport.target_socket` method. Do the same thing as
		basic method do, but also checks that the result address is IPv4 multicast address.

		:param config: beacon configuration
		:return: WIPV4SocketInfo
		"""
		target = WUDPNetworkNativeTransport.target_socket(self, config)
		if WNetworkIPV4.is_multicast(target.address()) is False:
			raise ValueError('IP multicast address not RFC compliant')
		return target

	@verify_type('paranoid', config=WConfig)
	def create_server_socket(self, config):
		""" Create server multicast socket. Socket will be joined to the multicast-group (same as it is
		specified in client configuration, same as client does)

		:param config: server configuration
		:return: socket.socket
		"""
		server_socket = WUDPNetworkNativeTransport.create_server_socket(self, config)

		group = socket.inet_aton(str(self.target_socket(config).address()))
		group_membership = struct.pack('4sL', group, socket.INADDR_ANY)
		server_socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, group_membership)
		return server_socket



# -*- coding: utf-8 -*-

import pytest

from wasp_general.crypto.hash import WHashGeneratorProto, WPyCryptoHashAdapter, WHash, WSHA512


def test_abstract():
	pytest.raises(TypeError, WHashGeneratorProto)
	pytest.raises(NotImplementedError, WHashGeneratorProto.update, None, b'')
	pytest.raises(NotImplementedError, WHashGeneratorProto.digest, None)
	pytest.raises(NotImplementedError, WHashGeneratorProto.generator_digest_size)
	pytest.raises(NotImplementedError, WHashGeneratorProto.generator_name)
	pytest.raises(NotImplementedError, WHashGeneratorProto.generator_family)
	pytest.raises(NotImplementedError, WHashGeneratorProto.new)


class TestWHashGeneratorProto:

	def test(self):
		class Dummy(WHashGeneratorProto):

			def update(self, data):
				pass

			def digest(self):
				return b'\x01\x02\x03\xff'

			@classmethod
			def generator_digest_size(cls):
				return 4

			@classmethod
			def generator_family(cls):
				return None

			@classmethod
			def generator_name(cls):
				return 'dummy'

			@classmethod
			def new(cls, data=None):
				return cls()

		assert(Dummy().hexdigest() == '010203FF')


class TestWPyCryptoHashAdapter:

	def test(self):

		assert(issubclass(WPyCryptoHashAdapter, WHashGeneratorProto) is True)

		class PyCryptoGenerator:

			def __init__(self):
				self.__digest = (b'\x00' * self.digest_size())

			def update(self, data):
				self.__digest = b''
				for i in range(min(len(data), self.digest_size())):
					self.__digest += bytes([data[i] - 1])
				for i in range(min(len(data), self.digest_size()), self.digest_size()):
					self.__digest += b'\x00'

			def digest(self):
				return self.__digest

			@classmethod
			def digest_size(cls):
				return 13

			@classmethod
			def new(cls, data=None):
				obj = cls()
				if data is not None:
					obj.update(data)
				return obj

		class BuggyPyCryptoAdapter(WPyCryptoHashAdapter):
			pass

		pytest.raises(ValueError, BuggyPyCryptoAdapter)
		pytest.raises(ValueError, BuggyPyCryptoAdapter.generator_digest_size)

		class PyCryptoAdapter(WPyCryptoHashAdapter):
			__pycrypto_cls__ = PyCryptoGenerator

		a = PyCryptoAdapter()
		assert(a.digest_size() == 13)
		assert(a.digest() == b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
		a.update(b'12345')
		assert(a.digest() == b'01234\x00\x00\x00\x00\x00\x00\x00\x00')

		assert(a.pycrypto() is not None)
		assert(isinstance(a.pycrypto(), PyCryptoGenerator) is True)

		pytest.raises(ValueError, PyCryptoAdapter.generator_name)
		PyCryptoAdapter.__generator_name__ = 1
		pytest.raises(TypeError, PyCryptoAdapter.generator_name)
		PyCryptoAdapter.__generator_name__ = 'foo'
		assert(PyCryptoAdapter.generator_name() == 'FOO')

		assert(PyCryptoAdapter.generator_family() is None)
		PyCryptoAdapter.__generator_family__ = 1
		pytest.raises(TypeError, PyCryptoAdapter.generator_family)
		PyCryptoAdapter.__generator_family__ = 'bar'
		assert(PyCryptoAdapter.generator_family() == 'BAR')


class TestWHash:

	def test(self):
		names = WHash.available_generators()
		assert('SHA1' in names)
		assert('SHA224' in names)
		assert('SHA256' in names)
		assert('SHA384' in names)
		assert('SHA512' in names)

		names = WHash.available_generators(name='SHA1')
		assert('SHA1' in names)

		sha1_result = b'^\x98\x99\xf2\x1f\xff\xa2\xee\x8b\x16\xbex\x03\x8d\xd8\x860^?\x82'
		assert(WHash.generator('SHA1').new(b'data to hash').digest() == sha1_result)

		assert(WHash.generator_by_digest('SHA', 64) == WSHA512)
		pytest.raises(ValueError, WHash.generator_by_digest, '???', 1)

		pytest.raises(ValueError, WHash.generator, '???')

		digests = WHash.available_digests('SHA')
		assert(digests == {20, 28, 32, 48, 64})



# -*- coding: utf-8 -*-
# wasp_general/network/web/template_response.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import io
from mako.template import Template
from mako.runtime import Context

from wasp_general.verify import verify_type

from wasp_general.template import WTemplate

from wasp_general.network.web.headers import WHTTPHeaders
from wasp_general.network.web.response import WWebResponse
from wasp_general.template import WTemplateRenderer


class WWebTemplateResponse(WTemplateRenderer, WWebResponse):

	@verify_type('paranoid', status=(int, None), template=WTemplate, context=(None, dict))
	@verify_type('paranoid', headers=(WHTTPHeaders, None))
	def __init__(self, template, context=None, status=None, headers=None):
		WTemplateRenderer.__init__(self, template, context=context)
		WWebResponse.__init__(self, status=status, headers=(headers if headers is not None else WHTTPHeaders()))

		if self.headers()['Content-Type'] is None:
			self.headers().add_headers('Content-Type', 'text/html')

	def response_data(self):
		return self.render()



# -*- coding: utf-8 -*-

from wasp_general.template import WTemplateText

from wasp_general.command.command import WCommand
from wasp_general.command.template_command import WTemplateResultCommand, WCommandResultTemplate


class TestWTemplateResultCommand:

	def test(self):
		template = WTemplateText('hello: ${var}')

		template_result_cmd = WTemplateResultCommand(template, 'test')
		assert(isinstance(template_result_cmd, WTemplateResultCommand) is True)
		assert(isinstance(template_result_cmd, WCommand) is True)
		assert(template_result_cmd.template() == template)
		assert(template_result_cmd.template_context() == {})

		template_result_cmd = WTemplateResultCommand(template, 'test', template_context={'var': 'world'})
		assert(template_result_cmd.template_context() == {'var': 'world'})

		result = template_result_cmd.result_template()
		assert(isinstance(result, WCommandResultTemplate) is True)
		assert(str(result) == 'hello: world')

		assert(template_result_cmd.match('test') is True)
		assert(template_result_cmd.match('test', 'foo') is False)

		result = template_result_cmd._exec('test')
		assert(isinstance(result, WCommandResultTemplate) is True)
		assert(str(result) == 'hello: world')



# -*- coding: utf-8 -*-

import pytest
from datetime import datetime

from wasp_general.task.dependency import WDependentTask, WTaskDependencyRegistryStorage, WTaskDependencyRegistry
from wasp_general.task.registry import WRegisteredTask
from wasp_general.task.base import WTask, WStoppableTask


class TestWDependentTask:

	def test_task(self):

		with pytest.raises(ValueError):
			class T1(metaclass=WDependentTask):
				pass

		with pytest.raises(TypeError):

			class FakeR:
				pass

			class T2(metaclass=WDependentTask):
				__registry__ = FakeR

		class R(WTaskDependencyRegistry):
			__registry_storage__ = WTaskDependencyRegistryStorage()
			last_task_result = None

		with pytest.raises(ValueError):
			class R2(WTaskDependencyRegistry):
				__registry_storage__ = WTaskDependencyRegistryStorage()
				__skip_none_registry_tag__ = False

			class T3(metaclass=WDependentTask):
				__registry__ = R2

		with pytest.raises(TypeError):
			class T4(metaclass=WDependentTask):
				__registry__ = R
				__registry_tag__ = 1

		class T5(WTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task.tag'

			def start(self):
				R.last_task_result = 'T5 call'

		assert(R.__registry_storage__.count() == 1)
		assert(R.last_task_result is None)
		assert(isinstance(T5.start_dependent_task(), T5) is True)
		assert(R.last_task_result == 'T5 call')


class TestWTaskDependencyRegistryStorage:

	def test_task(self):

		assert(WTaskDependencyRegistryStorage.__multiple_tasks_per_tag__ is False)

		registry_storage = WTaskDependencyRegistryStorage()

		class R(WTaskDependencyRegistry):
			__registry_storage__ = WTaskDependencyRegistryStorage()

		with pytest.raises(TypeError):
			class T1(WTask, metaclass=WRegisteredTask):
				__registry__ = R
				__registry_tag__ = 'task1'

		class T2(WTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task2'

			last_call_result = None

			def start(self):
				T2.last_call_result = datetime.now()

		assert(registry_storage.count() == 0)
		registry_storage.add(T2)
		assert (registry_storage.count() == 1)

		class T3(WStoppableTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task3'

			def start(self):
				pass

			def stop(self):
				pass

		class T4(WStoppableTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task4'
			__dependency__ = ['task3']

			def start(self):
				pass

			def stop(self):
				pass

		class T5(WStoppableTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task5'
			__dependency__ = ['task3', 'task4']

			def start(self):
				pass

			def stop(self):
				pass

		class T6(WStoppableTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task6'
			__dependency__ = ['task3', 'unknown_task']

			def start(self):
				pass

			def stop(self):
				pass

		class T7(WStoppableTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task7'
			__dependency__ = []

			def start(self):
				pass

			def stop(self):
				pass

		class T8(WStoppableTask, metaclass=WDependentTask):
			__registry__ = R
			__registry_tag__ = 'task8'
			__dependency__ = ['task5', 'task7']

			def start(self):
				pass

			def stop(self):
				pass

		registry_storage.add(T3)
		registry_storage.add(T4)
		registry_storage.add(T5)
		registry_storage.add(T6)
		registry_storage.add(T7)
		registry_storage.add(T8)
		assert(registry_storage.count() == 7)

		registry_storage.dependency_check(T2)
		registry_storage.dependency_check(T3)
		registry_storage.dependency_check(T4)
		registry_storage.dependency_check(T5)
		pytest.raises(RuntimeError, registry_storage.dependency_check, T6)
		registry_storage.dependency_check(T6, skip_unresolved=True)

		T5.__dependency__ = ['task4']
		T3.__dependency__ = ['task5']

		pytest.raises(RuntimeError, registry_storage.dependency_check, T3)
		pytest.raises(RuntimeError, registry_storage.dependency_check, T4)
		pytest.raises(RuntimeError, registry_storage.dependency_check, T5)

		T5.__dependency__ = ['task3', 'task4']
		T3.__dependency__ = []

		pytest.raises(RuntimeError, registry_storage.start_task, 'unknown tag')

		assert(registry_storage.started_tasks(task_registry_id='task2') is None)
		registry_storage.start_task('task2')
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task2'), T2) is True)

		last_result = T2.last_call_result
		registry_storage.start_task('task2')
		assert(last_result == T2.last_call_result)

		assert(registry_storage.started_tasks(task_registry_id='task3') is None)
		assert(registry_storage.started_tasks(task_registry_id='task4') is None)
		registry_storage.start_task('task4')
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task3'), T3) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task4'), T4) is True)

		assert(len(registry_storage.started_tasks()) == 3)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task3', task_cls=T3), T3) is True)
		assert(registry_storage.started_tasks(task_registry_id='task3', task_cls=T4) is None)
		assert(isinstance(registry_storage.started_tasks(task_cls=T3)[0], T3) is True)

		assert(registry_storage.started_tasks(task_registry_id='task6') is None)
		pytest.raises(RuntimeError, registry_storage.start_task, 'task6')
		registry_storage.start_task('task6', skip_unresolved=True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task6'), T6) is True)

		registry_storage.stop_task('unknown tag')
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task2'), T2) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task3'), T3) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task4'), T4) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task6'), T6) is True)

		registry_storage.stop_task('task3')
		assert(registry_storage.started_tasks(task_registry_id='task3') is None)
		assert(registry_storage.started_tasks(task_registry_id='task4') is None)
		assert(registry_storage.started_tasks(task_registry_id='task6') is None)

		T3.__dependency__ = ['unknown task']
		pytest.raises(RuntimeError, registry_storage.start_task, 'task5')
		T3.__dependency__ = []

		assert(registry_storage.started_tasks(task_registry_id='task3') is None)
		assert(registry_storage.started_tasks(task_registry_id='task4') is None)
		assert(registry_storage.started_tasks(task_registry_id='task5') is None)
		registry_storage.start_task('task5')
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task3'), T3) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task4'), T4) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task5'), T5) is True)
		registry_storage.stop_task('task3', stop_dependent=False)
		assert(registry_storage.started_tasks(task_registry_id='task3') is None)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task4'), T4) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task5'), T5) is True)

		registry_storage.start_task('task3')
		registry_storage.start_task('task8')
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task3'), T3) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task4'), T4) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task5'), T5) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task7'), T7) is True)
		assert(isinstance(registry_storage.started_tasks(task_registry_id='task8'), T8) is True)
		registry_storage.stop_task('task8', stop_requirements=True)
		assert(registry_storage.started_tasks(task_registry_id='task3') is None)
		assert(registry_storage.started_tasks(task_registry_id='task4') is None)
		assert(registry_storage.started_tasks(task_registry_id='task5') is None)
		assert(registry_storage.started_tasks(task_registry_id='task7') is None)
		assert(registry_storage.started_tasks(task_registry_id='task8') is None)


class TestWTaskDependencyRegistry:

	def test_registry(self):

		class R1(WTaskDependencyRegistry):
			pass

		with pytest.raises(ValueError):
			R1.registry_storage()

		with pytest.raises(TypeError):
			class FakeRS:
				pass
			R1.__registry_storage__ = FakeRS()
			R1.registry_storage()

		R1.__registry_storage__ = WTaskDependencyRegistryStorage()
		assert(R1.registry_storage() == R1.__registry_storage__)

		class T1(WTask, metaclass=WDependentTask):
			__registry__ = R1
			__registry_tag__ = 'task1'

			last_call_result = None

			def start(self):
				T1.last_call_result = 'T1.start'

			def stop(self):
				T1.last_call_result = 'T1.stop'

		class T2(WStoppableTask, metaclass=WDependentTask):
			__registry__ = R1
			__registry_tag__ = 'task2'

			last_call_result = None

			def start(self):
				T2.last_call_result = 'T2.start'

			def stop(self):
				T2.last_call_result = 'T2.stop'

		assert(R1.registry_storage().started_tasks(task_registry_id='task1') is None)
		assert(R1.registry_storage().started_tasks(task_registry_id='task2') is None)

		R1.start_task('task1')
		assert(isinstance(R1.registry_storage().started_tasks(task_registry_id='task1'), T1) is True)
		assert(T1.last_call_result == 'T1.start')
		assert(R1.registry_storage().started_tasks(task_registry_id='task2') is None)

		R1.stop_task('task1')
		assert(R1.registry_storage().started_tasks(task_registry_id='task1') is None)
		assert(T1.last_call_result == 'T1.start')
		assert(R1.registry_storage().started_tasks(task_registry_id='task2') is None)

		R1.start_task('task2')
		assert(R1.registry_storage().started_tasks(task_registry_id='task1') is None)
		assert(isinstance(R1.registry_storage().started_tasks(task_registry_id='task2'), T2) is True)
		assert(T2.last_call_result == 'T2.start')

		R1.stop_task('task2')
		assert(R1.registry_storage().started_tasks(task_registry_id='task1') is None)
		assert(R1.registry_storage().started_tasks(task_registry_id='task2') is None)
		assert(T2.last_call_result == 'T2.stop')

		R1.start_task('task1')
		R1.start_task('task2')
		assert(isinstance(R1.registry_storage().started_tasks(task_registry_id='task1'), T1) is True)
		assert(isinstance(R1.registry_storage().started_tasks(task_registry_id='task2'), T2) is True)

		R1.all_stop()
		assert(R1.registry_storage().started_tasks(task_registry_id='task1') is None)
		assert(R1.registry_storage().started_tasks(task_registry_id='task2') is None)



# -*- coding: utf-8 -*-
# wasp_general/crypto/rsa.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA as pyRSA

from wasp_general.verify import verify_type, verify_value
from wasp_general.crypto.hash import WHash


class WRSA:
	""" PyCrypto RSA-encryption wrapper
	"""

	wrapped_class = pyRSA.generate(bits=1024).__class__
	""" RSA wrapped class
	"""

	@staticmethod
	@verify_type(key_length=int)
	@verify_value(key_length=lambda x: ((x % 256) == 0) and x >= 1024)
	def generate_private(key_length=2048):
		""" Generate new private key (corresponding public key is included)

		:param key_length: same as bits argument in Crypto.PublicKey.RSA.generate function (it must be \
		a multiple of 256, and no smaller than 1024)
		:return: WRSA.wrapped_class
		"""
		return pyRSA.generate(key_length)

	@staticmethod
	@verify_type(key=wrapped_class, password=(str, bytes, None))
	def export_key(key, password=None):
		""" Export key in PEM-format

		:param key: key to export
		:param password: If it is not None, then result will be encrypt with given password. Suitable only \
		for private key. With public keys this argument does nothing
		:return: bytes
		"""

		return key.exportKey('PEM', passphrase=password)

	@staticmethod
	@verify_type(pem_text=(str, bytes), password=(str, bytes, None))
	def import_key(pem_text, password=None):
		""" Import key written in PEM-format

		:param pem_text: PEM data
		:param password: Password with witch PEM-data is encrypted
		:return: RSA.wrapped_class
		"""

		return pyRSA.importKey(pem_text, passphrase=password)

	@staticmethod
	@verify_type(private_key=wrapped_class)
	def generate_public(private_key):
		""" Get public key from private one

		:param private_key: source private key
		:return: RSA.wrapper_class
		"""

		return private_key.publickey()

	@staticmethod
	@verify_type(binary_chain=bytes, key=wrapped_class, sha_digest_size=int)
	@verify_value(sha_digest_size=lambda x: x in WHash.available_digests(family='SHA'))
	def encrypt(binary_chain, key, sha_digest_size=32):
		""" Encrypt data with key and PKCS1 OAEP protocol

		:param binary_chain: data to encrypt
		:param key: must be public key or private key with builtin public
		:param sha_digest_size: SHA digest size
		:return: bytes
		"""

		hash_generator = WHash.generator_by_digest('SHA', sha_digest_size).new()
		cipher = PKCS1_OAEP.new(key, hashAlgo=hash_generator)
		return cipher.encrypt(binary_chain)

	@staticmethod
	@verify_type(binary_chain=bytes, key=wrapped_class, sha_digest_size=int)
	@verify_value(sha_digest_size=lambda x: x in WHash.available_digests(family='SHA'))
	def decrypt(binary_chain, private_key, sha_digest_size=32):
		""" Decrypt data with key and PKCS1 OAEP protocol

		:param binary_chain: data to decrypt
		:param private_key: private key
		:param sha_digest_size: SHA digest size
		:return: bytes
		"""

		hash_generator = WHash.generator_by_digest('SHA', sha_digest_size).new()
		cipher = PKCS1_OAEP.new(private_key, hashAlgo=hash_generator)
		return cipher.decrypt(binary_chain)

	@staticmethod
	@verify_type(text=str, key=wrapped_class, sha_digest_size=int)
	@verify_value(sha_digest_size=lambda x: x in WHash.available_digests(family='SHA'))
	def string_encrypt(text, key, sha_digest_size=32):
		""" Encrypt text with given public key and PKCS1 OAEP protocol

		:param text: text to encrypt
		:param key: public key or private key with builtin public
		:param sha_digest_size: SHA digest size
		:return: bytes
		"""

		return WRSA.encrypt(text.encode(), key, sha_digest_size)

	@staticmethod
	@verify_type(binary_data=bytes, key=wrapped_class, text_encoding=(str, None), sha_digest_size=int)
	@verify_value(sha_digest_size=lambda x: x in WHash.available_digests(family='SHA'))
	def string_decrypt(binary_data, private_key, text_encoding=None, sha_digest_size=32):
		""" Decrypt binary data with given private key and PKCS1 OAEP protocol

		:param binary_data: data to decrypt
		:param private_key: private key
		:param text_encoding: source string encoding
		:param sha_digest_size: SHA digest size
		:return: str
		"""

		binary = WRSA.decrypt(binary_data, private_key, sha_digest_size)
		return binary.decode(text_encoding) if text_encoding is not None else binary.decode()



# -*- coding: utf-8 -*-
# wasp_general/task/scheduler/proto.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import uuid
from abc import ABCMeta, abstractmethod
from enum import Enum
from datetime import datetime, timezone

from wasp_general.verify import verify_type, verify_value

from wasp_general.task.thread import WThreadTask


# noinspection PyAbstractClass
class WScheduleTask(WThreadTask):
	""" Class represent task that may run by a scheduler
	Every schedule task must be able:
		- to be stopped at any time
		- to return its status (running or stopped)
		- to notify when task end (thread event)

	note: derived classes must implement :meth:`.WThreadTask.thread_started` and :meth:`.WThreadTask.thread_stopped`
	methods in order to be instantiable

	Each task instance has "unique" identifier
	"""

	__thread_name_prefix__ = 'ScheduledTask-'
	""" Thread name prefix
	"""

	@verify_type('paranoid', thread_join_timeout=(int, float, None))
	def __init__(self, thread_join_timeout=None):
		""" Create new task

		:param thread_join_timeout: same as thread_join_timeout in :meth:`.WThreadTask.__init__` method
		"""

		self.__uid = self.generate_uid()

		WThreadTask.__init__(
			self, thread_name=(self.__thread_name_prefix__ + str(self.__uid)), join_on_stop=True,
			ready_to_stop=True, thread_join_timeout=thread_join_timeout
		)

	def uid(self):
		return self.__uid

	@classmethod
	def generate_uid(cls):
		""" Return "random" "unique" identifier

		:return: UUID
		"""
		return uuid.uuid4()


class WScheduleRecord:
	""" This class specifies how :class:`.WScheduleTask` should run. It should be treated as scheduler record, that
	may not have execution time.

	:class:`.WScheduleRecord` has a policy, that describes what scheduler should do if it can not run this task
	at the specified moment. This policy is a recommendation for a scheduler and a scheduler can omit it if
	(for example) a scheduler queue is full. In any case, if this task is dropped (skipped) or postponed (moved to
	a queue of waiting tasks) correlated callback is called. "on_drop" callback is called for skipped tasks
	(it invokes via :meth:`.WScheduleRecord.task_dropped` method) and "on_wait" for postponed tasks (via
	:meth:`.WScheduleRecord.task_postponed` method)

	note: It is important that tasks with the same id (task_group_id) have the same postpone policy. If they do not
	have the same policy, then exception may be raised. No pre-checks are made to resolve this, because of
	unpredictable logic of different tasks from different sources
	"""
	# TODO: add policy that resolves concurrency of running tasks (like skipping tasks, that is already running)

	class PostponePolicy(Enum):
		""" Specifies what should be with this task if a scheduler won't be able to run it (like if the
		scheduler limit of running tasks is reached).
		"""
		wait = 1  # will stack every postponed task to execute them later (default)
		drop = 2  # drop this task if it can't be executed at the moment
		postpone_first = 3  # stack the first task and drop all the following tasks with the same task ID
		postpone_last = 4  # stack the last task and drop all the previous tasks with the same task ID

	@verify_type(task=WScheduleTask, task_group_id=(str, None))
	@verify_value(on_drop=lambda x: x is None or callable(x), on_wait=lambda x: x is None or callable(x))
	def __init__(self, task, policy=None, task_group_id=None, on_drop=None, on_wait=None):
		""" Create new schedule record

		:param task: task to run
		:param policy: postpone policy
		:param task_group_id: identifier that groups different scheduler records and single postpone policy
		:param on_drop: callback, that must be called if this task is skipped
		:param on_wait: callback, that must be called if this task is postponed
		"""

		if policy is not None and isinstance(policy, WScheduleRecord.PostponePolicy) is False:
			raise TypeError('Invalid policy object type')

		self.__task = task
		self.__policy = policy if policy is not None else WScheduleRecord.PostponePolicy.wait
		self.__task_group_id = task_group_id
		self.__on_drop = on_drop
		self.__on_wait = on_wait

	def task(self):
		""" Return task that should be run

		:return: WScheduleTask
		"""
		return self.__task

	def task_uid(self):
		""" Shortcut for self.task().uid()
		"""
		return self.task().uid()

	def policy(self):
		""" Return postpone policy

		:return: WScheduleRecord.PostponePolicy
		"""
		return self.__policy

	def task_group_id(self):
		""" Return task id

		:return: str or None

		see :meth:`.WScheduleRecord.__init__`
		"""
		return self.__task_group_id

	def task_postponed(self):
		""" Call a "on_wait" callback. This method is executed by a scheduler when it postpone this task

		:return: None
		"""
		if self.__on_wait is not None:
			return self.__on_wait()

	def task_dropped(self):
		""" Call a "on_drop" callback. This method is executed by a scheduler when it skip this task

		:return: None
		"""
		if self.__on_drop is not None:
			return self.__on_drop()


class WTaskSourceProto(metaclass=ABCMeta):
	""" Prototype for scheduler record generator. :class:`.WSchedulerServiceProto` doesn't have scheduler as set of
	records. Instead, a service uses this class as scheduler records holder and checks if it is time to execute
	them.
	"""

	@abstractmethod
	def has_records(self):
		""" Return records that should be run at the moment.


		:return: tuple of WScheduleRecord (tuple with one element at least) or None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def next_start(self):
		""" Return datetime when the next task should be executed.

		:return: datetime in UTC timezone
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def tasks_planned(self):
		""" Return number of records (tasks) that are planned to run

		:return: int
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def scheduler_service(self):
		""" Return associated scheduler service

		:return: WSchedulerServiceProto or None
		"""
		raise NotImplementedError('This method is abstract')


class WRunningRecordRegistryProto(metaclass=ABCMeta):
	""" This class describes a registry of running tasks. It executes a scheduler record
	(:class:`.WScheduleRecord`), creates and store the related records (:class:`.WScheduleRecord`), and watches
	that these tasks are running
	"""

	@abstractmethod
	@verify_type(schedule_record=WScheduleRecord)
	def exec(self, schedule_record):
		""" Execute the given scheduler record (no time checks are made at this method, task executes as is)

		:param schedule_record: record to execute

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def running_records(self):
		""" Return tuple of running tasks

		:return: tuple of WScheduleRecord
		"""
		raise NotImplementedError('This method is abstract')


class WSchedulerServiceProto(metaclass=ABCMeta):
	""" Represent a scheduler. A core of wasp_general.task.scheduler module
	"""

	@abstractmethod
	@verify_type(task_source=(WTaskSourceProto, None))
	def update(self, task_source=None):
		""" Update task sources information about next start. Update information for the specified source
		or for all of them

		:param task_source: if it is specified - then update information for this source only

		This method implementation must be thread-safe as different threads (different task source, different
		registries) may modify scheduler internal state.
		:return:
		"""
		raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.task.base import WTask, WStoppableTask, WTerminatableTask, WSyncTask


def test_abstract_classes():
	pytest.raises(TypeError, WTask)
	pytest.raises(NotImplementedError, WTask.start, None)

	pytest.raises(TypeError, WStoppableTask)
	pytest.raises(NotImplementedError, WStoppableTask.stop, None)

	pytest.raises(TypeError, WTerminatableTask)
	pytest.raises(NotImplementedError, WTerminatableTask.terminate, None)


class TestWSyncTask:

	def test_task(self):

		pytest.raises(TypeError, WSyncTask)

		class T1(WSyncTask):

			last_call_result = None

			def start(self):
				T1.last_call_result = 'T1.start'

		t1 = T1()
		assert(isinstance(t1, WTask) is True)
		assert(isinstance(t1, WStoppableTask) is True)

		assert(T1.last_call_result is None)
		t1.start()
		assert(T1.last_call_result == 'T1.start')

		t1.stop()
		assert(T1.last_call_result == 'T1.start')

		class T2(WSyncTask):

			last_call_result = None

			def start(self):
				T2.last_call_result = 'T2.start'

			def stop(self):
				T2.last_call_result = 'T2.stop'

		t2 = T2()
		t2.start()
		assert(T2.last_call_result == 'T2.start')
		t2.stop()
		assert(T2.last_call_result == 'T2.stop')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.command.command import WCommandProto, WCommand, WCommandSelector, WCommandPrioritizedSelector
from wasp_general.command.command import WCommandSet, WCommandAlias, WReduceCommand
from wasp_general.command.proto import WCommandResultProto
from wasp_general.command.result import WPlainCommandResult


def test_abstract():
	pytest.raises(TypeError, WCommandProto)
	pytest.raises(NotImplementedError, WCommandProto.match, None)
	pytest.raises(NotImplementedError, WCommandProto.exec, None)

	pytest.raises(TypeError, WCommand)
	pytest.raises(NotImplementedError, WCommand._exec, None)

	pytest.raises(TypeError, WCommandAlias)
	pytest.raises(NotImplementedError, WCommandAlias.mutate_command_tokens, None)


class TestWCommandProto:

	def test(self):
		split_result = WCommandProto.split_command('call "function 1" with\t0.1 "test words"')
		assert(split_result == ['call', 'function 1', 'with', '0.1', 'test words'])
		join_result = WCommandProto.join_tokens(*split_result)
		assert(join_result == "call 'function 1' with 0.1 'test words'")


class TestWCommand:

	class Command(WCommand):

		def _exec(self, *command_tokens):
			return WPlainCommandResult('OK')

	def test(self):
		command = TestWCommand.Command('create', 'world')
		assert(isinstance(command, WCommand) is True)
		assert(isinstance(command, WCommandProto) is True)

		assert(command.match('create', 'world', 'new world') is True)
		assert(command.match('update', 'world') is False)
		assert(command.match('create') is False)
		assert(TestWCommand.Command('create').match('create') is True)

		result = command.exec('create', 'world', '2.0')
		assert(isinstance(result, WCommandResultProto) is True)
		assert(str(result) == 'OK')

		pytest.raises(RuntimeError, command.exec, 'update')


class TestWCommandSelector:

	def test(self):
		create_cmd = TestWCommand.Command('create')
		create_world_cmd = TestWCommand.Command('create', 'world')

		command_selector = WCommandSelector()
		assert(len(command_selector) == 0)
		command_selector.add(create_cmd)
		assert(len(command_selector) == 1)
		command_selector.add(create_world_cmd)
		assert(len(command_selector) == 2)
		assert(command_selector.select('create') == create_cmd)
		assert(command_selector.select('create', 'world') == create_cmd)

		command_selector = WCommandSelector()
		command_selector.add(create_world_cmd)
		command_selector.add(create_cmd)
		assert(command_selector.select('create') == create_cmd)
		assert(command_selector.select('create', 'world') == create_world_cmd)

		assert(command_selector.select('update') is None)


class TestWCommandPrioritizedSelector:

	def test(self):
		create_cmd = TestWCommand.Command('create')
		create_world_cmd = TestWCommand.Command('create', 'world')

		command_selector = WCommandPrioritizedSelector()
		assert(len(command_selector) == 0)
		assert(isinstance(command_selector, WCommandPrioritizedSelector) is True)
		assert(isinstance(command_selector, WCommandSelector) is True)

		command_selector.add(create_cmd)
		assert(len(command_selector) == 1)
		command_selector.add(create_world_cmd)
		assert(len(command_selector) == 2)
		assert(command_selector.select('create') == create_cmd)
		assert(command_selector.select('create', 'world') == create_cmd)

		command_selector = WCommandPrioritizedSelector()
		command_selector.add(create_cmd)
		command_selector.add_prioritized(create_world_cmd, -1)
		assert(command_selector.select('create') == create_cmd)
		assert(command_selector.select('create', 'world') == create_world_cmd)


class TestWCommandSet:

	class Command(WCommand):

		def _exec(self, *command_tokens, **command_env):
			return WPlainCommandResult('context set', sec_var='1')

	def test(self):
		command_set = WCommandSet()
		assert(isinstance(command_set.commands(), WCommandSelector) is True)
		assert(isinstance(command_set.commands(), WCommandPrioritizedSelector) is False)

		create_cmd = TestWCommand.Command('create')
		create_world_cmd = TestWCommand.Command('create', 'world')
		command_selector = WCommandPrioritizedSelector()
		command_selector.add(create_cmd)
		command_selector.add_prioritized(create_world_cmd, -1)

		command_set = WCommandSet(command_selector)
		assert(command_set.commands() == command_selector)

		result = command_set.exec('create world 2.0')
		assert(isinstance(result, WCommandResultProto) is True)
		assert(str(result) == 'OK')

		pytest.raises(WCommandSet.NoCommandFound, command_set.exec, 'hello world')

		command_set = WCommandSet()
		assert(command_set.tracked_vars() == tuple())
		assert(command_set.has_var('sec_var') is False)

		command_set = WCommandSet(tracked_vars=('sec_var', ))
		assert(command_set.tracked_vars() == ('sec_var', ))
		assert(command_set.has_var('sec_var') is False)

		set_command = TestWCommandSet.Command('set')
		command_set.commands().add(set_command)
		assert(command_set.has_var('sec_var') is False)

		command_set.exec('set')
		assert(command_set.has_var('sec_var') is True)
		assert(command_set.var_value('sec_var') == '1')


class TestWCommandAlias:

	class Command(WCommandAlias):

		def mutate_command_tokens(self, *command_tokens):
			result = ['foo']
			result.extend(command_tokens)
			result.append('bar')
			return result

	def test(self):
		command_selector = WCommandSelector()
		cmd_alias = TestWCommandAlias.Command(command_selector)
		assert(isinstance(cmd_alias, TestWCommandAlias.Command) is True)
		assert(isinstance(cmd_alias, WCommandAlias) is True)
		assert(isinstance(cmd_alias, WCommandProto) is True)
		assert(cmd_alias.selector() == command_selector)
		assert(cmd_alias.match('test') is False)

		command_selector.add(TestWCommand.Command('test'))
		assert(cmd_alias.match('test') is False)

		command_selector.add(TestWCommand.Command('foo', 'test', 'bar'))
		assert(cmd_alias.match('test') is True)

		result = cmd_alias.exec('test')
		assert(isinstance(result, WCommandResultProto) is True)
		assert(str(result) == 'OK')

		def mutation_error_fn(*command_tokens):
			return

		cmd_alias.mutate_command_tokens = mutation_error_fn
		assert (cmd_alias.match('test') is False)
		pytest.raises(RuntimeError, cmd_alias.exec, 'test')


class TestWReduceCommand:

	def test(self):
		command_selector = WCommandSelector()
		pytest.raises(RuntimeError, WReduceCommand, command_selector)

		reduce_command = WReduceCommand(command_selector, 'section1', 'section2')
		assert(isinstance(reduce_command, WReduceCommand) is True)
		assert(isinstance(reduce_command, WCommandAlias) is True)
		assert(reduce_command.reduce_tokens() == ('section1', 'section2'))

		assert(reduce_command.mutate_command_tokens('hello') is None)
		assert(reduce_command.mutate_command_tokens('section1', 'hello', 'world') == ('hello', 'world'))
		assert(reduce_command.mutate_command_tokens('section2', 'hello', 'world') == ('hello', 'world'))



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/onion.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: write tests

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type

from wasp_general.network.messenger.proto import WMessengerOnionProto
from wasp_general.network.messenger.proto import WMessengerOnionLayerProto

from wasp_general.network.messenger.layers import WMessengerSimpleCastingLayer
from wasp_general.network.messenger.coders import WMessengerFixedModificationLayer, WMessengerEncodingLayer
from wasp_general.network.messenger.coders import WMessengerHexLayer, WMessengerBase64Layer, WMessengerAESLayer
from wasp_general.network.messenger.coders import WMessengerRSALayer

from wasp_general.network.messenger.packers import WMessengerJSONPacker

from wasp_general.network.messenger.composer import WMessengerComposerLayer

from wasp_general.network.messenger.transport import WMessengerSendAgentLayer, WMessengerSyncReceiveAgentLayer


class WMessengerOnion(WMessengerOnionProto):
	""" :class:`.WMessengerOnionProto` implementation. This class holds layers
	(:class:`WMessengerOnionLayerProto` class) that can be used for message processing.
	"""

	__builtin_layers__ = {x.name(): x for x in [
		WMessengerFixedModificationLayer(), WMessengerEncodingLayer(), WMessengerHexLayer(),
		WMessengerBase64Layer(), WMessengerAESLayer(), WMessengerRSALayer(), WMessengerJSONPacker(),
		WMessengerSendAgentLayer(), WMessengerSyncReceiveAgentLayer(), WMessengerComposerLayer(),
		WMessengerSimpleCastingLayer()
	]}
	""" Builtin layers
	"""

	@verify_type(layers=WMessengerOnionLayerProto)
	def __init__(self, *layers):
		""" Construct new onion

		:param layers: layers to store
		"""
		self.__layers = {}
		self.add_layers(*layers)

	def layers_names(self):
		""" :meth:`.WMessengerOnionProto.layer_names` method implementation.
		"""
		return list(self.__class__.__builtin_layers__.keys()) + list(self.__layers.keys())

	@verify_type(layer_name=str)
	def layer(self, layer_name):
		""" :meth:`.WMessengerOnionProto.layer` method implementation.
		"""
		if layer_name in self.__layers.keys():
			return self.__layers[layer_name]
		elif layer_name in self.__class__.__builtin_layers__:
			return self.__class__.__builtin_layers__[layer_name]

		raise RuntimeError('Invalid layer name')

	@verify_type(layer=WMessengerOnionLayerProto)
	def add_layers(self, *layers):
		""" Append given layers to this onion

		:param layers: layer to add
		:return: None
		"""
		for layer in layers:
			if layer.name() in self.__layers.keys():
				raise ValueError('Layer "%s" already exists' % layer.name())
			self.__layers[layer.name()] = layer



# -*- coding: utf-8 -*-
# wasp_general/mime.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.


# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import mimetypes
import magic
from threading import Lock


__mime_lock = Lock()


def mime_type(filename):
	""" Guess mime type for the given file name

	Note: this implementation uses python_magic package which is not thread-safe, as a workaround global lock is
	used for the ability to work in threaded environment

	:param filename: file name to guess
	:return: str
	"""
	# TODO: write lock-free mime_type function
	try:
		__mime_lock.acquire()

		extension = filename.split(".")
		extension = extension[len(extension) - 1]

		if extension == "woff2":
			return "application/font-woff2"
		if extension == "css":
			return "text/css"

		m = magic.from_file(filename, mime=True)
		m = m.decode() if isinstance(m, bytes) else m  # compatibility fix, some versions return bytes some - str
		if m == "text/plain":
			guessed_type = mimetypes.guess_type(filename)[0]  # for js-detection
			if guessed_type:
				return guessed_type
		return m
	finally:
		__mime_lock.release()



# -*- coding: utf-8 -*-

import pytest
import time

from wasp_general.task.base import WStoppableTask, WTerminatableTask, WTask
from wasp_general.task.thread import WThreadTask, WThreadCustomTask, WThreadJoiningTimeoutError, WPollingThreadTask
from wasp_general.task.thread import WThreadedTaskChain


class TestWThreadTask:

	def test_init(self):
		assert(issubclass(WThreadTask, WStoppableTask) is True)
		assert(issubclass(WThreadTask, WTerminatableTask) is False)
		pytest.raises(TypeError, WThreadTask)
		pytest.raises(NotImplementedError, WThreadTask.thread_started, None)
		pytest.raises(NotImplementedError, WThreadTask.thread_stopped, None)

		class T(WThreadTask):

			def thread_started(self):
				pass

			def thread_stopped(self):
				pass

		T()

	def test_task(self):

		class FastTask(WThreadTask):

			__thread_join_timeout__ = 3

			call_stack = []

			def thread_started(self):
				FastTask.call_stack.append('FastTask.start')

			def thread_stopped(self):
				FastTask.call_stack.append('FastTask.stop')

		class SlowTask(FastTask):

			sleep_time = 0.5

			def thread_started(self):
				FastTask.start(self)
				time.sleep(SlowTask.sleep_time)

		t = FastTask()
		assert(t.thread() is None)
		assert(t.join_timeout() == FastTask.__thread_join_timeout__)
		assert(t.start_event().is_set() is False)
		assert(t.stop_event().is_set() is False)
		assert(t.ready_event() is None)

		t.start()
		assert(t.thread() is not None)
		assert(t.thread().name != 'custom thread name')
		assert(t.start_event().is_set() is True)
		assert(t.stop_event().is_set() is False)
		assert(t.ready_event() is None)

		t.stop()
		assert(t.thread() is None)
		assert(t.start_event().is_set() is False)
		assert(t.stop_event().is_set() is True)
		assert(FastTask.call_stack == ['FastTask.start', 'FastTask.stop'])
		assert(t.ready_event() is None)

		FastTask.__thread_join_timeout__ = 2
		t = FastTask(thread_name='custom thread name')
		assert(t.join_timeout() == FastTask.__thread_join_timeout__)
		t.start()
		assert(t.thread().name == 'custom thread name')
		t.stop()

		FastTask.__thread_name__ = 'class thread name'
		t = FastTask()
		t.start()
		assert(t.thread().name == 'class thread name')
		t.stop()

		t = FastTask(thread_join_timeout=4)
		assert(t.join_timeout() != FastTask.__thread_join_timeout__)
		assert(t.join_timeout() == 4)

		t = SlowTask(thread_join_timeout=0.01)
		t.start()
		assert(t.stop_event().is_set() is False)
		pytest.raises(WThreadJoiningTimeoutError, t.stop)
		assert(t.stop_event().is_set() is True)
		t.thread().join(SlowTask.sleep_time)
		t.close_thread()
		assert(t.thread() is None)
		assert(t.stop_event().is_set() is True)

		FastTask.call_stack = []
		t = FastTask(join_on_stop=False)
		assert(t.thread() is None)
		t.start()
		assert(t.thread() is not None)
		t.stop()
		assert(t.thread() is not None)
		assert(FastTask.call_stack == ['FastTask.start', 'FastTask.stop'])
		t.thread().join()
		t.close_thread()
		assert(t.thread() is None)

		slow_task = SlowTask(ready_to_stop=True)
		assert(slow_task.ready_event().is_set() is False)

		slow_task.start()
		assert(slow_task.ready_event().is_set() is False)
		assert(slow_task.stop_event().is_set() is False)

		slow_task.ready_event().wait(SlowTask.sleep_time * 2)
		assert(slow_task.ready_event().is_set() is True)
		assert(slow_task.stop_event().is_set() is False)

		slow_task.stop()
		assert(slow_task.ready_event().is_set() is True)
		assert(slow_task.stop_event().is_set() is True)

		t = FastTask()
		raised_exception = []

		def test_exception():
			raise KeyError('Test exception')

		def handle_exception(exc):
			raised_exception.append(exc.__class__)

		t.thread_started = test_exception
		assert (t.exception_event().is_set() is False)
		t.start()
		t.exception_event().wait(FastTask.__thread_join_timeout__)
		time.sleep(0.1)
		assert(t.exception_event().is_set() is True)
		t.stop()
		assert (t.exception_event().is_set() is True)

		assert(raised_exception == [])
		t.thread_exception = handle_exception
		t.start()
		t.exception_event().wait(FastTask.__thread_join_timeout__)
		time.sleep(0.1)
		assert(raised_exception == [KeyError])
		t.stop()
		assert(raised_exception == [KeyError])


class TestWThreadCustomTask:

	__call_stack__ = []

	class Task(WTask):

		def start(self):
			TestWThreadCustomTask.__call_stack__.append('Task::start')

	class StoppableTask(WStoppableTask):

		def start(self):
			TestWThreadCustomTask.__call_stack__.append('StoppableTask::start')

		def stop(self):
			TestWThreadCustomTask.__call_stack__.append('StoppableTask::stop')

	def test(self):
		task = TestWThreadCustomTask.Task()
		threaded_task = WThreadCustomTask(task)

		assert(isinstance(threaded_task, WThreadCustomTask) is True)
		assert(isinstance(threaded_task, WThreadTask) is True)

		TestWThreadCustomTask.__call_stack__.clear()
		assert(TestWThreadCustomTask.__call_stack__ == [])

		threaded_task.start()
		assert(TestWThreadCustomTask.__call_stack__ == ['Task::start'])

		threaded_task.stop()
		assert(TestWThreadCustomTask.__call_stack__ == ['Task::start'])

		task = TestWThreadCustomTask.StoppableTask()
		threaded_task = WThreadCustomTask(task)

		TestWThreadCustomTask.__call_stack__.clear()
		assert(TestWThreadCustomTask.__call_stack__ == [])

		threaded_task.start()
		assert(TestWThreadCustomTask.__call_stack__ == ['StoppableTask::start'])

		threaded_task.stop()
		assert(TestWThreadCustomTask.__call_stack__ == ['StoppableTask::start', 'StoppableTask::stop'])


class TestWPollingThreadTask:

	class Task(WPollingThreadTask):

		call_stack = []

		def _polling_iteration(self):
			TestWPollingThreadTask.Task.call_stack.append('Task iteration')

		def thread_stopped(self):
			TestWPollingThreadTask.Task.call_stack.append('Task stop')

	def test(self):
		pytest.raises(TypeError, WPollingThreadTask)
		pytest.raises(NotImplementedError, WPollingThreadTask._polling_iteration, None)
		pytest.raises(NotImplementedError, WPollingThreadTask.thread_stopped, None)

		task = TestWPollingThreadTask.Task()
		assert(isinstance(task, WPollingThreadTask) is True)
		assert(isinstance(task, WThreadTask) is True)
		assert(task.polling_timeout() == WPollingThreadTask.__thread_polling_timeout__)

		timeout = WPollingThreadTask.__thread_polling_timeout__ + 2
		task = TestWPollingThreadTask.Task(polling_timeout=timeout)
		assert(task.polling_timeout() == timeout)

		timeout = 0.001
		task = TestWPollingThreadTask.Task(polling_timeout=timeout)
		task.start()
		time.sleep(timeout * 5)
		task.stop()
		assert(TestWPollingThreadTask.Task.call_stack[:4] == [
			'Task iteration', 'Task iteration', 'Task iteration', 'Task iteration'
		])
		assert (TestWPollingThreadTask.Task.call_stack[-1] == 'Task stop')

		TestWPollingThreadTask.Task.call_stack.clear()
		TestWPollingThreadTask.Task.__thread_polling_timeout__ = 0.1
		task = TestWPollingThreadTask.Task()
		assert(task.polling_timeout() == TestWPollingThreadTask.Task.__thread_polling_timeout__)
		task.start()
		time.sleep(timeout * 2)
		task.stop()
		assert(len(TestWPollingThreadTask.Task.call_stack[:-1]) <= 3)
		for call in TestWPollingThreadTask.Task.call_stack[:-1]:
			assert(call == 'Task iteration')
		assert (TestWPollingThreadTask.Task.call_stack[-1] == 'Task stop')


class TestWThreadedTaskChain:

	def test(self):
		polling_timeout = 0.001
		chain = WThreadedTaskChain(polling_timeout=polling_timeout)
		assert(isinstance(chain, WThreadedTaskChain) is True)
		assert(isinstance(chain, WPollingThreadTask) is True)

		assert (chain.ready_event().is_set() is False)
		chain.start()
		time.sleep(polling_timeout * 5)
		assert(chain.ready_event().is_set() is True)
		chain.stop()

		class Task(WThreadTask):

			call_trace = []

			def __init__(self, sleep_timeout, call_trace_id, ready_to_stop=True):
				WThreadTask.__init__(self, ready_to_stop=ready_to_stop)
				self.__sleep_timeout = sleep_timeout
				self.__trace_id = call_trace_id

			def thread_started(self):
				time.sleep(self.__sleep_timeout)
				Task.call_trace.append(self.__trace_id)
				self.stop_event().set()

			def thread_stopped(self):
				pass

		pytest.raises(ValueError, WThreadedTaskChain, Task(0, 'task', ready_to_stop=False))

		task1_sleep_timeout = 0.1
		task1 = Task(task1_sleep_timeout, 'task1')

		task2_sleep_timeout = 0.001
		task2 = Task(task2_sleep_timeout, 'task2')

		assert(Task.call_trace == [])
		task1.start()
		task2.start()
		time.sleep(task1_sleep_timeout * 5)
		task1.stop()
		task2.stop()
		assert(Task.call_trace == ['task2', 'task1'])

		Task.call_trace.clear()
		chain = WThreadedTaskChain(task1, task2, polling_timeout=polling_timeout)
		assert(Task.call_trace == [])
		chain.start()
		time.sleep(task1_sleep_timeout * 5)
		assert(Task.call_trace == ['task1', 'task2'])
		chain.stop()

		def test_exception():
			print('RAISE EXCEPTION!')
			raise KeyError('Test exception')
		task2.thread_started = test_exception

		task2_raised_exception = []

		def task2_handle_exception(exc):
			print('TASK CAUGHT EXC: ' + str(exc))
			task2_raised_exception.append(exc.__class__)

		task2.thread_exception = task2_handle_exception

		chain_raised_exception = []

		def chain_handle_exception(exc):
			print('CHAIN CAUGHT EXC: ' + str(exc))
			chain_raised_exception.append(exc.__class__)

		assert(task2_raised_exception == [])
		assert(chain_raised_exception == [])
		chain = WThreadedTaskChain(task1, task2, polling_timeout=polling_timeout)
		chain.thread_exception = chain_handle_exception
		chain.start()
		chain.exception_event().wait(polling_timeout * 5)
		time.sleep(0.5)
		assert(task2_raised_exception == [KeyError])
		assert(chain_raised_exception == [RuntimeError])
		chain.stop()
		assert(task2_raised_exception == [KeyError])
		assert(chain_raised_exception == [RuntimeError])

		assert(Task.call_trace == ['task1', 'task2', 'task1'])



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.web.response import WWebResponse
from wasp_general.network.web.headers import WHTTPHeaders


class TestWWebResponse:

	def test_response(self):
		assert(isinstance(WWebResponse(), WWebResponse) is True)
		assert(WWebResponse().__pushed_responses__() == tuple())
		assert(WWebResponse(status=301).status() == 301)
		assert(WWebResponse(response_data=b'data!').response_data() == b'data!')
		assert(WWebResponse(headers=WHTTPHeaders(headers1='value')).headers()['headers1'] == ('value',))

		response = WWebResponse()
		pushed_response1 = WWebResponse()
		pushed_response2 = WWebResponse()
		response.__push__(pushed_response1, pushed_response2)
		assert(response.__pushed_responses__() == (pushed_response1, pushed_response2))



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.web.debug import WWebDebugInfo
from wasp_general.network.web.proto import WWebSessionProto
from wasp_general.network.web.request import WWebRequest
from wasp_general.network.web.response import WWebResponse


def test_abstract():

	class Session(WWebSessionProto):

		def client_address(self):
			pass

		def server_address(self):
			pass

		def protocol(self):
			pass

		def protocol_version(self):
			pass

		def read_request(self):
			pass

		def write_response(self, request, response):
			pass

		def session_close(self):
			pass

	pytest.raises(TypeError, WWebDebugInfo)
	pytest.raises(NotImplementedError, WWebDebugInfo.session_id, None)
	pytest.raises(
		NotImplementedError, WWebDebugInfo.request, None, None, WWebRequest(Session(), 'GET', '/'), '2', 'https'
	)
	pytest.raises(NotImplementedError, WWebDebugInfo.response, None, None, WWebResponse())
	pytest.raises(NotImplementedError, WWebDebugInfo.target_route, None, None, None)
	pytest.raises(NotImplementedError, WWebDebugInfo.exception, None, None, ValueError(''))
	pytest.raises(NotImplementedError, WWebDebugInfo.finalize, None, None)



# -*- coding: utf-8 -*-
# wasp_general/network/service.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: write tests for the code
# TODO: update docs

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from datetime import timedelta
from abc import abstractmethod, ABCMeta
from threading import Event
import queue

from zmq import Context as ZMQContext
from zmq.eventloop.ioloop import IOLoop
from zmq.eventloop.zmqstream import ZMQStream

from wasp_general.verify import verify_type, verify_subclass
from wasp_general.config import WConfig

from wasp_general.network.transport import WNetworkNativeTransportProto


class WIOLoopServiceHandler(metaclass=ABCMeta):
	""" Represent service (or service client) handler that works with tornado IOLoop and do the work.
	"""

	@abstractmethod
	@verify_type(io_loop=IOLoop)
	def setup_handler(self, io_loop):
		""" Set up this handler with the specified IOLoop

		:param io_loop: service (or service client) loop to use with
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	def loop_stopped(self):
		""" Method is called when previously set up loop was stopped

		:return: None
		"""
		pass


class WIOLoopService(metaclass=ABCMeta):
	""" Represent service (or service client) that works over tornado IOLoop
	"""

	@verify_type(handler=WIOLoopServiceHandler, loop=(IOLoop, None), timeout=(int, None))
	def __init__(self, handler, loop=None, timeout=None):
		""" Create new service (or service client)

		:param handler: handler that do the work
		:param loop: loop to use (or None for internal one)
		:param timeout: timeout after which this loop will be stopped
		"""
		self.__loop = IOLoop() if loop is None else loop
		self.__handler = handler
		self.__timeout = timeout

	def loop(self):
		""" Return service loop object

		:return: IOLoop
		"""
		return self.__loop

	def handler(self):
		""" Return service handler

		:return: WIOLoopServiceHandler
		"""
		return self.__handler

	def timeout(self):
		""" Return service timeout. (None for endless loop)

		:return: int or None
		"""
		return self.__timeout

	def start(self):
		""" Set up handler and start loop

		:return: None
		"""
		timeout = self.timeout()
		if timeout is not None and timeout > 0:
			self.__loop.add_timeout(timedelta(0, timeout), self.stop)
		self.handler().setup_handler(self.loop())
		self.loop().start()
		self.handler().loop_stopped()

	def stop(self):
		""" Stop loop

		:return: None
		"""
		self.loop().stop()


class WNativeSocketIOHandler(metaclass=ABCMeta):
	""" Handler prototype for loops that work with :class:`.WNetworkNativeTransportProto` transports. It is used by
	:class:`.WBasicNativeSocketHandler` handler and do all the work for this class.
	"""

	@abstractmethod
	def handler_fn(self, fd, event):
		""" Process (handle) specified event

		:param fd: integer file descriptor or a file-like object with a fileno() method
		:param event: IOLoop event
		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WBasicNativeSocketHandler(WIOLoopServiceHandler, metaclass=ABCMeta):
	""" Basic :class:`.WIOLoopServiceHandler` implementation. Since some :class:`.WNetworkNativeTransportProto`
	methods are required :class:`.WConfig` object, then that kind of object is required for this class
	instantiation
	"""

	@verify_type(transport=WNetworkNativeTransportProto, config=WConfig, io_handler=WNativeSocketIOHandler)
	def __init__(self, transport, config, io_handler):
		""" Create new socket handler

		:param transport: transport to use
		:param config: configuration to be used with transport
		:param io_handler: handler that do the real work
		"""
		WIOLoopServiceHandler.__init__(self)
		self.__transport = transport
		self.__config = config
		self.__io_handler = io_handler

	def transport(self):
		""" Return currently used transport
		:return: WNetworkNativeTransportProto
		"""
		return self.__transport

	def config(self):
		""" Return handler configuration

		:return: WConfig
		"""
		return self.__config

	def io_handler(self):
		""" Return IO-handler

		:return: WNativeSocketIOHandler
		"""
		return self.__io_handler


class WNativeSocketDirectIOHandler(WNativeSocketIOHandler, metaclass=ABCMeta):
	""" This type of IO-handler has access to low-level socket object
	"""

	def __init__(self):
		""" Create new IO-handler

		"""
		WNativeSocketIOHandler.__init__(self)
		self.__transport_socket = None

	def transport_socket(self, new_socket=None):
		""" Save and/or return currently used socket object

		:param new_socket: new socket to save
		:return: socket object (any type, None if socket wasn't set)
		"""
		if new_socket is not None:
			self.__transport_socket = new_socket
		return self.__transport_socket


class WNativeSocketHandler(WBasicNativeSocketHandler):
	""" Enhanced variant of :class:`.WBasicNativeSocketHandler` class. This class support 'server_mode' flag and
	is capable to set up the specified IO-handler with :class:`.WIOLoopService` service
	"""

	@verify_type('paranoid', transport=WNetworkNativeTransportProto, config=WConfig)
	@verify_type('paranoid', io_handler=WNativeSocketIOHandler)
	@verify_type(server_mode=bool)
	def __init__(self, transport, config, io_handler, server_mode):
		""" Create new loop-handler

		:param transport: transport to use
		:param config: configuration to use (in the most cases it is used by transport object only)
		:param io_handler: io-handler to use
		:param server_mode: set 'server_mode' flag for correct transport configuration
		"""
		WBasicNativeSocketHandler.__init__(self, transport, config, io_handler)
		self.__server_mode = server_mode

	def server_mode(self):
		""" Return current mode. True if this handler works as a server, otherwise - False

		:return: bool
		"""
		return self.__server_mode

	@verify_type(io_loop=IOLoop)
	def setup_handler(self, io_loop):
		""" :meth:`.WIOLoopServiceHandler.setup_handler` implementation.
		If :class:`.WNativeSocketDirectIOHandler` is used as a io-handler, then socket object is saved
		to this handler before loop starting

		:param io_loop: io_loop to use
		:return: None
		"""

		if self.server_mode() is True:
			s = self.transport().server_socket(self.config())
		else:
			s = self.transport().client_socket(self.config())

		io_handler = self.io_handler()
		if isinstance(io_handler, WNativeSocketDirectIOHandler) is True:
			io_handler.transport_socket(s)
		io_loop.add_handler(s.fileno(), io_handler.handler_fn, io_loop.READ)

	def loop_stopped(self):
		""" Terminate socket connection because of stopping loop

		:return: None
		"""
		transport = self.transport()
		if self.server_mode() is True:
			transport.close_server_socket(self.config())
		else:
			transport.close_client_socket(self.config())


class WZMQHandler(WIOLoopServiceHandler, metaclass=ABCMeta):

	class SocketOption:

		def __init__(self, name, value):
			self.name = name
			self.value = value

	class SetupAgent:

		@verify_type(socket_type=int, connection=str)
		def __init__(self, socket_type, connection, *socket_options):
			self.__socket_type = socket_type
			self.__connection = connection
			for option in socket_options:
				if isinstance(option, WZMQHandler.SocketOption) is False:
					raise TypeError('Invalid socket option')
			self.__socket_options = socket_options if len(socket_options) > 0 else tuple()

		def socket_type(self):
			return self.__socket_type

		def connection(self):
			return self.__connection

		def socket_options(self):
			return self.__socket_options

		def create_socket(self, handler):
			if isinstance(handler, WZMQHandler) is False:
				raise TypeError('Invalid handler type')
			context = handler.context()
			for option in self.socket_options():
				context.setsockopt(option.name, option.value)
			return context.socket(self.socket_type())

		@verify_type(io_loop=IOLoop)
		def setup_handler(self, handler, io_loop):
			if isinstance(handler, WZMQHandler) is False:
				raise TypeError('Invalid handler type')
			s = self.create_socket(handler)
			return ZMQStream(s, io_loop=io_loop)

		def setup_receiver(self, handler, receive_agent):
			if isinstance(handler, WZMQHandler) is False:
				raise TypeError('Invalid handler type')

			if isinstance(receive_agent, WZMQHandler.ReceiveAgent) is False:
				raise TypeError('Invalid receive agent type')

			stream = handler.stream()
			if stream is None:
				raise RuntimeError("Handler stream wasn't set")

			def callback(msg):
				receive_agent.on_receive(handler, msg)

			receive_agent.setup_receiver(handler)
			stream.on_recv(callback)

	class BindSetupAgent(SetupAgent):

		def create_socket(self, handler):
			socket = WZMQHandler.SetupAgent.create_socket(self, handler)
			socket.bind(self.connection())
			return socket

	class ConnectSetupAgent(SetupAgent):

		def create_socket(self, handler):
			socket = WZMQHandler.SetupAgent.create_socket(self, handler)
			socket.connect(self.connection())
			return socket

	class SendAgent:

		@verify_type(data=bytes)
		def send(self, handler, data):
			if isinstance(handler, WZMQHandler) is False:
				raise TypeError('Invalid handler type')

			stream = handler.stream()
			if stream is None:
				raise RuntimeError("Handler stream wasn't set")

			stream.send(data)
			stream.flush()

	class ReceiveAgent:

		@abstractmethod
		def on_receive(self, handler, msg):
			raise NotImplementedError('This method is abstract')

		def setup_receiver(self, handler):
			pass

	@verify_type(context=(ZMQContext, None))
	def __init__(self, context=None):
		self.__context = context if context is not None else ZMQContext()
		self.__stream = None
		self.__setup_agent = None
		self.__receive_agent = None

	def context(self):
		return self.__context

	def stream(self):
		return self.__stream

	def setup_agent(self):
		return self.__setup_agent

	def receive_agent(self):
		return self.__receive_agent

	def configure(self, setup_agent, receive_agent=None):
		if self.__setup_agent is not None:
			raise RuntimeError('Unable to configure handler multiple time')
		self.__setup_agent = setup_agent
		self.__receive_agent = receive_agent

	@verify_type(io_loop=IOLoop)
	def setup_handler(self, io_loop):
		setup_agent = self.setup_agent()
		self.__stream = setup_agent.setup_handler(self, io_loop)
		receive_agent = self.receive_agent()
		if receive_agent is not None:
			setup_agent.setup_receiver(self, receive_agent)


class WZMQService(WIOLoopService):

	@verify_type('paranoid', setup_agent=WZMQHandler.SetupAgent, loop=(IOLoop, None))
	@verify_type('paranoid', receive_agent=(WZMQHandler.ReceiveAgent, None), timeout=(int, None))
	@verify_type(handler=(WZMQHandler, None))
	def __init__(self, setup_agent, loop=None, handler=None, receive_agent=None, timeout=None):
		if handler is None:
			handler = WZMQHandler()
		handler.configure(setup_agent, receive_agent)
		WIOLoopService.__init__(self, handler, loop=loop, timeout=timeout)

	def discard_queue_messages(self):
		""" Sometimes it is necessary to drop undelivered messages. These messages may be stored in different
		caches, for example in a zmq socket queue. With different zmq flags we can tweak zmq sockets and
		contexts no to keep those messages. But inside ZMQStream class there is a queue that can not be
		cleaned other way then the way it does in this method. So yes, it is dirty to access protected
		members, and yes it can be broken at any moment. And yes without correct locking procedure there
		is a possibility of unpredicted behaviour. But still - there is no other way to drop undelivered
		messages

		Discussion of the problem: https://github.com/zeromq/pyzmq/issues/1095

		:return: None
		"""
		zmq_stream_queue = self.handler().stream()._send_queue
		while not zmq_stream_queue.empty():
			try:
				zmq_stream_queue.get(False)
			except queue.Empty:
				continue
			zmq_stream_queue.task_done()


class WLoglessIOLoop(IOLoop):

	def _setup_logging(self):
		pass


class WZMQSyncAgent(WZMQHandler.ReceiveAgent):

	@verify_type(timeout=(int, float, None))
	def __init__(self, timeout=None):
		self.__timeout = timeout
		self.__threaded_event = Event()
		self.__data = None
		self.__handler = None

	def timeout(self):
		return self.__timeout

	def event(self):
		return self.__threaded_event

	def data(self):
		data = self.__data
		self.__data = None
		self.__threaded_event.clear()
		return data

	def handler(self):
		return self.__handler

	def on_receive(self, handler, msg):
		if self.__data is not None:
			raise RuntimeError('Multiple responses for a single request')
		self.__data = msg
		self.__threaded_event.set()

	@verify_type(handler=WZMQHandler)
	def setup_receiver(self, handler):
		if self.__handler is None:
			self.__handler = handler
		else:
			raise RuntimeError('Unable to setup receive agent multiple times')

	def receive(self):
		status = self.event().wait(self.timeout())
		if status is True:
			return self.data()

		raise TimeoutError('Request timeout reached')



# -*- coding: utf-8 -*-

import pytest
import os
from inspect import isfunction

from wasp_general.verify import Verifier, TypeVerifier, SubclassVerifier, ValueVerifier
from wasp_general.verify import verify_type, verify_subclass, verify_value


@pytest.fixture
def verifier_env(request):
	def fin():
		if Verifier.__default_environment_var__ in os.environ:
			del os.environ[Verifier.__default_environment_var__]
	request.addfinalizer(fin)


def fn_name_checker():
	pass


class FNameChecker:
	@staticmethod
	def foo():
		pass

	@classmethod
	def bar(cls):
		pass

	def zzz(self):
		pass


@pytest.mark.usefixtures('verifier_env')
class TestVerifier:

	def test_decorate_disabled(self):

		pytest.raises(KeyError, 'os.environ[Verifier.__default_environment_var__]')

		assert(Verifier().decorate_disabled() is False)
		assert(Verifier('test_tag1').decorate_disabled() is False)
		assert(Verifier('test_tag1', 'test_tag2').decorate_disabled() is False)

		os.environ[Verifier.__default_environment_var__] = 'foo:bar:test_tag1:value3'
		assert(Verifier().decorate_disabled() is False)
		assert(Verifier('test_tag1').decorate_disabled() is True)
		assert(Verifier('test_tag1', 'test_tag2').decorate_disabled() is False)

		os.environ[Verifier.__default_environment_var__] = 'foo:bar:test_tag1:value3:test_tag2'
		assert(Verifier().decorate_disabled() is False)
		assert(Verifier('test_tag1').decorate_disabled() is True)
		assert(Verifier('test_tag1', 'test_tag2').decorate_disabled() is True)

	def test_check(self):
		check = Verifier().check(None, '', lambda x: None)
		assert(isfunction(check) is True)

	def test_decorator(self):
		pytest.raises(KeyError, 'os.environ[Verifier.__default_environment_var__]')

		def foo(a, b, c, d=None, **kwargs):
			'''
			multi-line
			bla-bla docstring
			'''
			pass

		verifier = Verifier()

		assert(verifier.decorator()(foo) != foo)
		verifier.decorate_disabled = lambda: True
		assert(verifier.decorator()(foo) == foo)
		verifier = Verifier()

		def exc():
			raise TypeError('text exception')

		default_check = lambda x: None
		exc_check = lambda x: exc() if x == 3 else None
		a_spec = lambda x: 111
		verifier.check = lambda s, n, f: exc_check if n == 'a' or n == 'd' or n == 'e' else default_check
		decorated_foo = verifier.decorator(a=a_spec, c=1, d=1, e=1)(foo)

		decorated_foo(1, 2, 3, d=4)
		pytest.raises(TypeError, decorated_foo, 3, 2, 3, d=4)
		pytest.raises(TypeError, decorated_foo, 1, 2, 3, d=3)
		decorated_foo(1, 2, 3, d=4, e=5)
		pytest.raises(TypeError, decorated_foo, 1, 2, 3, d=4, e=3)

	def test_function_name(self):

		assert(Verifier.function_name(FNameChecker.foo) == 'FNameChecker.foo')
		assert(Verifier.function_name(FNameChecker.bar) == 'FNameChecker.bar')
		assert(Verifier.function_name(FNameChecker.zzz) == 'FNameChecker.zzz')

		c = FNameChecker()
		assert(Verifier.function_name(c.foo) == 'FNameChecker.foo')
		assert(Verifier.function_name(c.bar) == 'FNameChecker.bar')
		assert(Verifier.function_name(c.zzz) == 'FNameChecker.zzz')


class TestTypeVerifier:

	def test_check(self):

		def foo(a, b, c, d=None, **kwargs):
			'''
			multi-line
			bla-bla docstring
			'''
			pass

		verifier = TypeVerifier()
		pytest.raises(RuntimeError, 'verifier.decorator(a=None)(foo)')
		pytest.raises(RuntimeError, 'verifier.decorator(a=(str, None, 1))(foo)')

		decorated_foo = verifier.decorator(a=int, b=(str, None), c=[str, int], d=(str, int, None), e=float)(foo)
		decorated_foo(1, None, 'f')
		decorated_foo(1, None, 1, d='o')
		decorated_foo(1, None, 'o', d=5, e=.1)

		pytest.raises(TypeError, "decorated_foo('b', None, 'f')")
		pytest.raises(TypeError, "decorated_foo(1, 4, 'o')")
		pytest.raises(TypeError, "decorated_foo(1, None, None)")
		pytest.raises(TypeError, "decorated_foo(1, None, 'o', d=.1)")
		pytest.raises(TypeError, "decorated_foo(1, None, 'b', e='a')")

		def foo(*args):
			pass

		decorated_foo = verifier.decorator(args=int)(foo)
		decorated_foo()
		decorated_foo(1, 2, 3)
		pytest.raises(TypeError, decorated_foo, 0.1)
		pytest.raises(TypeError, decorated_foo, 1, 0.1, 4)

		def foo(x, *args):
			pass

		decorated_foo = verifier.decorator(args=(int, float))(foo)
		decorated_foo('')
		decorated_foo('', 0.1, 2, 3)
		pytest.raises(TypeError, decorated_foo, '', '')


class TestSubclassVerifier:

	def test_check(self):

		def foo(a, b, c, d=None, **kwargs):
			'''
			multi-line
			bla-bla docstring
			'''
			pass

		verifier = SubclassVerifier()
		pytest.raises(RuntimeError, 'verifier.decorator(a=None)(foo)')
		pytest.raises(RuntimeError, 'verifier.decorator(a=(str, None, 1))(foo)')

		class A:
			pass

		class B(A):
			pass

		class C(A):
			pass

		class D:
			pass

		decorated_foo = verifier.decorator(a=A, c=(B, C), d=[B, None], e=C)(foo)
		decorated_foo(A, None, B)
		decorated_foo(B, None, B)
		decorated_foo(A, None, B, d=B)
		decorated_foo(A, None, C, d=B)
		decorated_foo(A, None, B, d=None, e=C)

		pytest.raises(TypeError, "decorated_foo(D, None, B)")
		pytest.raises(TypeError, "decorated_foo(A, None, A)")
		pytest.raises(TypeError, "decorated_foo(A, None, None)")
		pytest.raises(TypeError, "decorated_foo(A, None, B, d=C)")
		pytest.raises(TypeError, "decorated_foo(A, None, B, d=B, e=A)")


class TestValueVerifier:

	def test_check(self):

		def foo(a, b, c, d=None, **kwargs):
			'''
			multi-line
			bla-bla docstring
			'''
			pass

		verifier = ValueVerifier()
		pytest.raises(RuntimeError, 'verifier.decorator(a=1)(foo)')
		pytest.raises(RuntimeError, 'verifier.decorator(a=(1,))(foo)')

		decorated_foo = verifier.decorator(
			a=(lambda x: x > 5, lambda x: x < 10),
			c=lambda x: x[:3] == 'foo', d=lambda x: x is not None, e=lambda x: x == 1
		)(foo)
		decorated_foo(6, 1, 'foo-asdaads', d=7)
		decorated_foo(6, 1, 'foo-asdaads', d='sss')
		decorated_foo(6, 1, 'foo-asdaads', d=0.1, e=1)

		pytest.raises(ValueError, decorated_foo, 6, None, 'foo-aaa')
		pytest.raises(ValueError, decorated_foo, 5, None, 'foo-aaa', d=5)
		pytest.raises(ValueError, decorated_foo, 10, None, 'foo-aaa', d=5)
		pytest.raises(ValueError, decorated_foo, 6, None, 'foo-aaa', d=5, e=7)


def test_verify():

	class A:
		a = 'foo'

	@verify_type(a=int, b=str, d=(int, None), e=float)
	@verify_subclass(c=A)
	@verify_value(a=(lambda x: x > 5, lambda x: x < 10), c=lambda x: x.a == 'foo', d=lambda x: x is None or x < 0)
	def foo(a, b, c, d=None, **kwargs):
		'''
		multi-line
		bla-bla docstring
		'''
		pass

	foo(6, 'bar', A)
	foo(7, 'bar', A, d=-1)
	foo(6, 'bar', A, e=0.1)

	pytest.raises(TypeError, foo, 'foo', 'bar', A)
	pytest.raises(TypeError, foo, 6, 'bar', int)
	pytest.raises(ValueError, foo, 3, 'bar', A)
	A.a = 'bar'
	pytest.raises(ValueError, foo, 6, 'bar', A)
	A.a = 'foo'
	pytest.raises(ValueError, foo, 6, 'bar', A, d=1)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.crypto.kdf import WPBKDF2
from wasp_general.crypto.hmac import WHMAC


class TestWPBKDF2:

	def test(self):
		kdf = WPBKDF2(b'very-very-very much secret key')
		assert(kdf.salt() is not None)
		assert(len(kdf.derived_key()) == WPBKDF2.__default_derived_key_length__)

		kdf = WPBKDF2(
			'very-very-very strong password', salt=b'public salt value', derived_key_length=20, iterations_count=2000,
			hmac=WHMAC.hmac('HMAC-SHA256')
		)
		assert(kdf.derived_key() == b'\xfaup\xf6\x04\xaf\xb1z\xbc*\xa5\xafy\n\xb01\x06\xb5\x0b\x94')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.task.thread import WThreadTask

from wasp_general.network.beacon.beacon import WNetworkServerBeacon

from wasp_general.network.beacon.task import WNetworkBeaconTask


class TestNetworkBeaconTask:

	def test_task(self):
		task = WNetworkBeaconTask()
		assert(isinstance(task, WThreadTask) is True)
		assert(isinstance(task, WNetworkServerBeacon) is True)
		assert(task.thread() is None)
		task.start()
		assert(task.thread() is not None)
		task.stop()
		assert(task.thread() is None)



# -*- coding: utf-8 -*-
# wasp_general/network/web/tornado.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import weakref
from tornado.web import RequestHandler

from wasp_general.network.web.proto import WWebResponseProto
from wasp_general.verify import verify_type
from wasp_general.network.web.headers import WHTTPHeaders
from wasp_general.network.web.session import WWebSessionAdapter
from wasp_general.network.web.request import WWebRequest
from wasp_general.network.web.cookies import WHTTPCookieJar


class WTornadoRequestHandler(RequestHandler):
	"""
	According to http://www.tornadoweb.org/en/stable/web.html, requesthandlers are not thread safe

	"""

	def __init__(self, wasp_web_service, application, request, **kwargs):
		RequestHandler.__init__(self, application, request, **kwargs)
		self.__sessions = {}
		self.__wasp_web_service = wasp_web_service

	def compute_etag(self):
		pass

	def __handle_request(self):

		fileno = self.request.connection.stream.socket.fileno()

		def close_session():
			if fileno in self.__sessions.keys():
				del self.__sessions[fileno]
			else:
				pass
				# very strange

		if fileno in self.__sessions.keys():
			session = self.__sessions[fileno]
		else:
			session = WTornadoSessionAdapter(self, close_session)
			self.__sessions[fileno] = session

		self.__wasp_web_service.process_request(session)

	def get(self, *args, **kwargs):
		self.__handle_request()

	def head(self, *args, **kwargs):
		self.__handle_request()

	def post(self, *args, **kwargs):
		self.__handle_request()

	def delete(self, *args, **kwargs):
		self.__handle_request()

	def patch(self, *args, **kwargs):
		self.__handle_request()

	def put(self, *args, **kwargs):
		self.__handle_request()

	def options(self, *args, **kwargs):
		self.__handle_request()

	@classmethod
	def __handler__(self, wasp_web_service):
		class Hanlder(WTornadoRequestHandler):
			def __init__(self, application, request, **kwargs):
				WTornadoRequestHandler.__init__(self, wasp_web_service, application, request, **kwargs)
		return Hanlder


class WTornadoSessionAdapter(WWebSessionAdapter):

	def __init__(self, request_handler, cleanup_handler):

		def weakref_handler(socket_ref):
			self.session_close()

		self.__request_handler = request_handler
		self.__socket_ref = weakref.ref(request_handler.request.connection.stream.socket, weakref_handler)
		self.__protocol_version = request_handler.request.version[len('HTTP/'):]
		self.__protocol = request_handler.request.protocol
		self.__cleanup_handler = cleanup_handler

	def accepted_socket(self):
		return self.__socket_ref()

	def protocol_version(self):
		return self.__protocol_version

	def protocol(self):
		return self.__protocol

	def read_request(self):
		handler_headers = self.__request_handler.request.headers
		headers = WHTTPHeaders()
		for header_name in handler_headers.keys():
			headers.add_headers(header_name, *handler_headers.get_list(header_name))

		for cookie in WHTTPCookieJar.import_simple_cookie(self.__request_handler.cookies):
			headers.set_cookie_jar().add_cookie(cookie)

		request = WWebRequest(
			self, self.__request_handler.request.method, self.__request_handler.request.path,
			headers=headers.ro()
		)
		#! todo: fix
		#request.set_request_body()
		return request

	@verify_type(request=WWebRequest, reponse=WWebResponseProto, pushed_responses=WWebResponseProto)
	def write_response(self, request, response, *pushed_responses):
		status = response.status()
		headers = response.headers()
		response_data = response.response_data()

		if status is not None:
			self.__request_handler.set_status(status)

		if headers is not None:
			headers = headers.switch_name_style(self.protocol_version())
			for header_name in headers.headers():
				for header_value in headers[header_name]:
					self.__request_handler.add_header(header_name, header_value)

			content_type = headers.content_type()
			if content_type is not None:
				self.__request_handler.set_header(headers.normalize_name('Content-Type'), content_type)
			elif response_data is not None:
				self.__request_handler.set_header(
					headers.normalize_name('Content-Type'), 'application/octet-stream'
				)

			for cookie in headers.set_cookie_jar():
				self.__request_handler.set_cookie(
					cookie.name(), cookie.value(), **cookie.attrs_as_dict()
				)

		if response_data is not None:
			self.__request_handler.write(response_data)

	def session_close(self):
		self.__cleanup_handler()



# -*- coding: utf-8 -*-
# wasp_general/command/result.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general. If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type
from wasp_general.command.proto import WCommandResultProto


# noinspection PyAbstractClass
class WCommandEnv(WCommandResultProto):

	def __init__(self, **command_env):
		WCommandResultProto.__init__(self)
		self.__command_env = command_env

	def environment(self):
		return self.__command_env.copy()


class WPlainCommandResult(WCommandEnv):

	@verify_type(result=str)
	def __init__(self, result, **command_env):
		WCommandEnv.__init__(self, **command_env)
		self.__result = result

	def __str__(self):
		return self.__result

	@classmethod
	def error(cls, message, **command_env):
		return WPlainCommandResult('Error. ' + message, **command_env)


class WExceptionResult(WCommandEnv):

	@verify_type(message=str, exception=Exception)
	def __init__(self, message, exc, traceback, **command_env):
		WCommandEnv.__init__(self, **command_env)
		self.__message = message
		self.__exception = str(exc)
		self.__traceback = traceback

	def message(self):
		return self.__message

	def exception(self):
		return self.__exception

	def traceback(self):
		return self.__traceback

	def __str__(self):
		return (
			'%s\nException was raised. %s\n%s' %
			(self.message(), self.exception(), self.traceback())
		)



# -*- coding: utf-8 -*-
# wasp_general/task/base.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type


class WTask(metaclass=ABCMeta):
	""" Basic task prototype. Must implement the only thing - to start
	"""

	@abstractmethod
	def start(self):
		""" Start this task

		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WStoppableTask(WTask):
	""" Task that can be stopped (graceful shutdown)
	"""

	@abstractmethod
	def stop(self):
		""" Stop this task (graceful shutdown)

		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WTerminatableTask(WStoppableTask):
	""" Task that can be terminated (rough shutdown)
	"""

	@abstractmethod
	def terminate(self):
		""" Terminate this task (rough shutdown)

		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WSyncTask(WStoppableTask, metaclass=ABCMeta):
	""" This class is some kind of declaration, that the following task is executed in foreground.
	"""

	def __init__(self):
		WStoppableTask.__init__(self)

	def stop(self):
		""" Stop this task. This implementation does nothing.

		:return: None
		"""
		pass



# -*- coding: utf-8 -*-
# wasp_general/task/thread_tracker.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
import traceback
from enum import Enum

from wasp_general.verify import verify_type, verify_value

from wasp_general.task.thread import WThreadTask
from wasp_general.task.scheduler.proto import WScheduleTask, WScheduleRecord
from wasp_general.thread import WCriticalResource
from wasp_general.datetime import utc_datetime


class WTrackerEvents(Enum):
	""" Possible tracking events
	"""
	start = 1  # start
	stop = 2  # normal stop
	termination = 3  # termination stop
	exception = 4  # unhandled exception stop
	wait = 5  # task has been postponed (scheduler event)
	drop = 6  # task has been dropped (scheduler event)


class WThreadTrackerInfoStorageProto(metaclass=ABCMeta):
	""" Prototype for a storage that keeps thread task events like start event, normal stop, termination, raised
	unhandled exceptions) or scheduler events
	"""

	@abstractmethod
	@verify_type(task=WThreadTask, event_details=(str, None))
	def register_start(self, task, event_details=None):
		""" Store start event

		:param task: task that is starting
		:param event_details: (optional) event details - any kind of data related to the given task and start \
		event

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(task=WThreadTask, event_details=(str, None))
	def register_stop(self, task, event_details=None):
		""" Store stop event

		:param task: task that stopped
		:param event_details: (optional) event details - any kind of data related to the given task and stop \
		event

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(task=WThreadTask, event_details=(str, None))
	def register_termination(self, task, event_details=None):
		""" Store termination event

		:param task: task that was terminated
		:param event_details: (optional) event details - any kind of data related to the given task and \
		termination event

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(task=WThreadTask, raised_exception=Exception, exception_details=str, event_details=(str, None))
	def register_exception(self, task, raised_exception, exception_details, event_details=None):
		""" Store exception event

		:param task: task that was terminated by unhandled exception
		:param raised_exception: unhandled exception
		:param exception_details: any kind of data related to the raised exception
		:param event_details: (optional) event details - any kind of data related to the given task and
		exception event

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(task=WThreadTask, event_details=(str, None))
	def register_wait(self, task, event_details=None):
		""" Store event of task postponing (this event is used in a scheduler classes)

		:param task: task that was postponed
		:param event_details: (optional) event details - any kind of data related to the given task and
		postponing event

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(task=WThreadTask, event_details=(str, None))
	def register_drop(self, task, event_details=None):
		""" Store event of task drop (this event is used in a scheduler classes)

		:param task: task that was dropped
		:param event_details: (optional) event details - any kind of data related to the given task and
		event of task drop

		:return: None
		"""
		raise NotImplementedError('This method is abstract')


# noinspection PyAbstractClass
class WThreadTracker(WThreadTask):
	""" Threaded task that may register its events (start event, normal stop fact, task termination fact,
	unhandled exceptions)

	:note: Since there is an extra work that should be done, this class may be inappropriate for low-latency
	situation. Also, registering events should be done quickly because of this task joining timeout.
	"""

	@verify_type('paranoid', thread_name=(str, None), join_on_stop=bool, thread_join_timeout=(int, float, None))
	@verify_type(tracker_storage=(WThreadTrackerInfoStorageProto, None), track_start=bool, track_stop=bool)
	@verify_type(track_termination=bool, track_exception=bool)
	def __init__(
		self, tracker_storage=None, thread_name=None, join_on_stop=True, thread_join_timeout=None,
		track_start=True, track_stop=True, track_termination=True, track_exception=True
	):
		""" Create new tracker

		:param tracker_storage: storage that is used for registering eventd
		:param thread_name: same as 'thread_name' in :meth:`.WThreadTask.__init__`
		:param join_on_stop: same as 'join_on_stop' in :meth:`.WThreadTask.__init__`
		:param thread_join_timeout: same as 'thread_join_timeout' in :meth:`.WThreadTask.__init__`
		:param track_start: whether to register start event of this task or not
		:param track_stop: whether to register stop event of this task or not
		:param track_termination: whether to register termination event of this task or not
		:param track_exception: whether to register unhandled exception event or not
		"""
		WThreadTask.__init__(
			self,
			thread_name=thread_name,
			join_on_stop=join_on_stop,
			ready_to_stop=True,
			thread_join_timeout=thread_join_timeout
		)
		self.__tracker = tracker_storage
		self.__track_start = track_start
		self.__track_stop = track_stop
		self.__track_termination = track_termination
		self.__track_exception = track_exception

	def tracker_storage(self):
		""" Return linked storage

		:return: WThreadTrackerInfoStorageProto or None
		"""
		return self.__tracker

	def track_start(self):
		""" Return True if this task is tracking "start-event" otherwise - False

		:return: bool
		"""
		return self.__track_start

	def track_stop(self):
		""" Return True if this task is tracking "stop-event" otherwise - False

		:return: bool
		"""
		return self.__track_stop

	def track_termination(self):
		""" Return True if this task is tracking "termination-event" otherwise - False

		:return: bool
		"""
		return self.__track_termination

	def track_exception(self):
		""" Return True if this task is tracking unhandled exception event otherwise - False

		:return: bool
		"""
		return self.__track_exception

	# noinspection PyMethodMayBeStatic
	@verify_type(event=WTrackerEvents)
	def event_details(self, event):
		""" Return task details that should be registered with a tracker storage

		:param event: source event that requested details

		:return: str or None
		"""
		return None

	def start(self):
		""" :meth:`.WThreadTask.start` implementation. Register (if required) start event by a tracker storage

		:return: None
		"""
		tracker = self.tracker_storage()
		if tracker is not None and self.track_start() is True:
			details = self.event_details(WTrackerEvents.start)
			tracker.register_start(self, event_details=details)
		WThreadTask.start(self)

	def thread_stopped(self):
		""" :meth:`.WThreadTask.thread_stopped` implementation. Register (if required) stop and termination
		event by a tracker storage

		:return: None
		"""
		tracker = self.tracker_storage()
		if tracker is not None:
			try:
				if self.ready_event().is_set() is True:
					if self.track_stop() is True:
						details = self.event_details(WTrackerEvents.stop)
						tracker.register_stop(self, event_details=details)
				elif self.exception_event().is_set() is False:
					if self.track_termination() is True:
						details = self.event_details(WTrackerEvents.termination)
						tracker.register_termination(self, event_details=details)
			except Exception as e:
				self.thread_tracker_exception(e)

	@verify_type(raised_exception=Exception)
	def thread_exception(self, raised_exception):
		""" :meth:`.WThreadTask.thread_exception` implementation. Register (if required) unhandled exception
		event by a tracker storage

		:param raised_exception: unhandled exception

		:return: None
		"""
		tracker = self.tracker_storage()
		if tracker is not None:
			try:
				if self.track_exception() is True:
					details = self.event_details(WTrackerEvents.exception)
					tracker.register_exception(
						self,
						raised_exception,
						traceback.format_exc(),
						event_details=details
					)
			except Exception as e:
				self.thread_tracker_exception(e)

	# noinspection PyMethodMayBeStatic
	@verify_type(raised_exception=Exception)
	def thread_tracker_exception(self, raised_exception):
		""" Method is called whenever an exception is raised during registering a event

		:param raised_exception: raised exception

		:return: None
		"""
		print('Thread tracker execution was stopped by the exception. Exception: %s' % str(raised_exception))
		print('Traceback:')
		print(traceback.format_exc())


class WSimpleTrackerStorage(WCriticalResource, WThreadTrackerInfoStorageProto):
	""" Simple :class:`.WThreadTrackerInfoStorageProto` implementation which stores events in a operation memory
	"""

	__critical_section_timeout__ = 1
	""" Timeout for capturing a lock for critical sections
	"""

	class Record:
		""" General record of single event
		"""

		@verify_type(record_type=WTrackerEvents, thread_task=WThreadTask)
		@verify_type(event_details=(str, None))
		def __init__(self, record_type, thread_task, event_details=None):
			""" Create new record

			:param record_type: tracking event
			:param thread_task: original task
			:param event_details: task details
			"""
			self.record_type = record_type
			self.thread_task = thread_task
			self.event_details = event_details
			self.registered_at = utc_datetime()

	class ExceptionRecord(Record):
		""" Record for unhandled exception
		"""

		@verify_type('paranoid', task=WThreadTask, event_details=(str, None))
		@verify_type(exception=Exception, exception_details=str)
		def __init__(self, task, exception, exception_details, event_details=None):
			WSimpleTrackerStorage.Record.__init__(
				self, WTrackerEvents.exception, task, event_details=event_details
			)
			self.exception = exception
			self.exception_details = exception_details

	@verify_type(records_limit=(int, None), record_start=bool, record_stop=bool, record_termination=bool)
	@verify_type(record_exception=bool, record_wait=bool, record_drop=bool)
	def __init__(
		self, records_limit=None, record_start=True, record_stop=True, record_termination=True,
		record_exception=True, record_wait=True, record_drop=True
	):
		""" Create new storage

		:param records_limit: number of records to keep (if record limit is reached - new record will
		overwrite the oldest one)
		:param record_start:  whether to keep start events or not
		:param record_stop: whether to keep normal stop events or not
		:param record_termination: whether to keep termination stop events or not
		:param record_exception: whether to keep unhandled exceptions events or not
		:param record_wait: whether to keep postponing events or not
		:param record_drop: whether to keep drop events or not
		"""
		WCriticalResource.__init__(self)
		WThreadTrackerInfoStorageProto.__init__(self)
		self.__limit = records_limit
		self.__registry = []
		self.__record_start = record_start
		self.__record_stop = record_stop
		self.__record_termination = record_termination
		self.__record_exception = record_exception
		self.__record_wait = record_wait
		self.__record_drop = record_drop

	def record_limit(self):
		""" Return maximum number of records to keep

		:return: int or None (for no limit)
		"""
		return self.__limit

	def record_start(self):
		""" Return True if this storage is saving start events, otherwise - False

		:return: bool
		"""
		return self.__record_start

	def record_stop(self):
		""" Return True if this storage is saving normal stop events, otherwise - False

		:return: bool
		"""
		return self.__record_stop

	def record_termination(self):
		""" Return True if this storage is saving termination stop events, otherwise - False

		:return: bool
		"""
		return self.__record_termination

	def record_exception(self):
		""" Return True if this storage is saving unhandled exceptions events, otherwise - False

		:return: bool
		"""
		return self.__record_exception

	def record_wait(self):
		""" Return True if this storage is saving postponing events, otherwise - False

		:return: bool
		"""
		return self.__record_wait

	def record_drop(self):
		""" Return True if this storage is saving dropping events, otherwise - False

		:return: bool
		"""
		return self.__record_drop

	def register_start(self, task, event_details=None):
		""" :meth:`.WSimpleTrackerStorage.register_start` method implementation
		"""
		if self.record_start() is True:
			record_type = WTrackerEvents.start
			record = WSimpleTrackerStorage.Record(record_type, task, event_details=event_details)
			self.__store_record(record)

	@verify_type(task=WThreadTask, details=(str, None))
	def register_stop(self, task, event_details=None):
		""" :meth:`.WSimpleTrackerStorage.register_stop` method implementation
		"""
		if self.record_stop() is True:
			record_type = WTrackerEvents.stop
			record = WSimpleTrackerStorage.Record(record_type, task, event_details=event_details)
			self.__store_record(record)

	@verify_type(task=WThreadTask, details=(str, None))
	def register_termination(self, task, event_details=None):
		""" :meth:`.WSimpleTrackerStorage.register_termination` method implementation
		"""
		if self.record_termination() is True:
			record_type = WTrackerEvents.termination
			record = WSimpleTrackerStorage.Record(record_type, task, event_details=event_details)
			self.__store_record(record)

	@verify_type(task=WThreadTask, raised_exception=Exception, exception_details=str, details=(str, None))
	def register_exception(self, task, raised_exception, exception_details, event_details=None):
		""" :meth:`.WSimpleTrackerStorage.register_exception` method implementation
		"""
		if self.record_exception() is True:
			record = WSimpleTrackerStorage.ExceptionRecord(
				task, raised_exception, exception_details, event_details=event_details
			)
			self.__store_record(record)

	@verify_type(task=WThreadTask, details=(str, None))
	def register_wait(self, task, event_details=None):
		""" :meth:`.WSimpleTrackerStorage.register_wait` method implementation
		"""
		if self.record_wait() is True:
			record_type = WTrackerEvents.wait
			record = WSimpleTrackerStorage.Record(record_type, task, event_details=event_details)
			self.__store_record(record)

	@verify_type(task=WThreadTask, details=(str, None))
	def register_drop(self, task, event_details=None):
		""" :meth:`.WSimpleTrackerStorage.register_drop` method implementation
		"""
		if self.record_drop() is True:
			record_type = WTrackerEvents.drop
			record = WSimpleTrackerStorage.Record(record_type, task, event_details=event_details)
			self.__store_record(record)

	@WCriticalResource.critical_section(timeout=__critical_section_timeout__)
	def __store_record(self, record):
		""" Save record in a internal storage

		:param record: record to save

		:return: None
		"""
		if isinstance(record, WSimpleTrackerStorage.Record) is False:
			raise TypeError('Invalid record type was')
		limit = self.record_limit()
		if limit is not None and len(self.__registry) >= limit:
			self.__registry.pop(0)
		self.__registry.append(record)

	@WCriticalResource.critical_section(timeout=__critical_section_timeout__)
	def __registry_copy(self):
		""" Return copy of tracked events

		:return: list of WSimpleTrackerStorage.Record
		"""
		return self.__registry.copy()

	def __iter__(self):
		""" Iterate over registered events (WSimpleTrackerStorage.Record). The newest record will be yield
		the first

		:return: generator
		"""
		registry = self.__registry_copy()
		while len(registry) > 0:
			yield registry.pop(-1)

	@verify_type(requested_events=WTrackerEvents)
	def last_record(self, task_uid, *requested_events):
		""" Search over registered :class:`.WScheduleTask` instances and return the last record that matches
		search criteria.

		:param task_uid: uid of :class:`.WScheduleTask` instance
		:param requested_events: target events types

		:return: WSimpleTrackerStorage.Record or None
		"""
		for record in self:
			if isinstance(record.thread_task, WScheduleTask) is False:
				continue
			if record.thread_task.uid() == task_uid:
				if len(requested_events) == 0 or record.record_type in requested_events:
					return record


class WScheduleRecordTracker(WScheduleRecord):
	""" Schedule record that may register scheduler events related to a scheduled task like 'postponing' event
	and 'drop' event.

	:note: Since there is an extra work that should be done, this class may be inappropriate for low-latency
	situation. Also, registering events should be done quickly because of this task joining timeout.
	"""

	@verify_type('paranoid', task=WScheduleTask, task_group_id=(str, None))
	@verify_value('paranoid', on_drop=lambda x: x is None or callable(x))
	@verify_value('paranoid', on_wait=lambda x: x is None or callable(x))
	@verify_type(task=WThreadTracker)
	@verify_type(track_wait=bool, track_drop=bool)
	def __init__(
		self, task, policy=None, task_group_id=None, on_drop=None, on_wait=None,
		track_wait=True, track_drop=True
	):
		""" Create new schedule record, that may track schedule events

		:param task: same as task in :meth:`.WScheduleRecord.__init__` except it must be \
		:class:`.WThreadTracker` instance
		:param policy: same as policy in :meth:`.WScheduleRecord.__init__`
		:param task_group_id: same as task_group_id in :meth:`.WScheduleRecord.__init__`
		:param on_drop: same as on_drop in :meth:`.WScheduleRecord.__init__`
		:param on_wait: same as on_wait in :meth:`.WScheduleRecord.__init__`
		:param track_wait: whether to register postponing event of this record or not
		:param track_drop: whether to register drop event of this record or not
		"""
		WScheduleRecord.__init__(
			self, task, policy=policy, task_group_id=task_group_id, on_drop=on_drop, on_wait=on_wait
		)
		self.__track_wait = track_wait
		self.__track_drop = track_drop

	def track_wait(self):
		""" Return True if this task is tracking "postponing" event otherwise - False

		:return: bool
		"""
		return self.__track_wait

	def track_drop(self):
		""" Return True if this task is tracking "drop" event otherwise - False

		:return: bool
		"""
		return self.__track_drop

	def task_postponed(self):
		""" Track (if required) postponing event and do the same job as :meth:`.WScheduleRecord.task_postponed`
		method do

		:return: None
		"""
		tracker = self.task().tracker_storage()
		if tracker is not None and self.track_wait() is True:
			details = self.task().event_details(WTrackerEvents.wait)
			tracker.register_wait(self.task(), event_details=details)
		WScheduleRecord.task_postponed(self)

	def task_dropped(self):
		""" Track (if required) drop event and do the same job as :meth:`.WScheduleRecord.task_dropped`
		method do

		:return: None
		"""
		tracker = self.task().tracker_storage()
		if tracker is not None and self.track_drop() is True:
			details = self.task().event_details(WTrackerEvents.drop)
			tracker.register_drop(self.task(), event_details=details)
		WScheduleRecord.task_dropped(self)



# -*- coding: utf-8 -*-
# wasp_general/command/proto.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod


class WCommandResultProto(metaclass=ABCMeta):

	@abstractmethod
	def __str__(self):
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def environment(self):
		raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# wasp_general/cache.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import weakref
from decorator import decorator
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_value, verify_type


class WCacheStorage(metaclass=ABCMeta):
	""" Abstract class for cache storage
	"""

	class CacheMissedException(Exception):
		""" Exception is raised in :meth:`.WCacheStorage.get_result` and derived methods as an error for
		accessing cache record that doesn't exist
		"""
		pass

	class CacheEntry:
		""" Cache request result, is used in :meth:`.WCacheStorage.get_cache` to determine if there is a
		cached value and what that value is.
		"""

		@verify_type(has_value=bool)
		def __init__(self, has_value=False, cached_value=None):
			""" Create new request result

			:param has_value: defines whether there is a cached value (True) or not (False)
			:param cached_value: defines cached value, this parameter should be used only if
			has_value is True
			"""
			self.has_value = has_value
			self.cached_value = cached_value

	@abstractmethod
	@verify_value(decorated_function=lambda x: callable(x))
	def put(self, result, decorated_function, *args, **kwargs):
		""" Save (or replace) result for given function

		:param result: result to be saved
		:param decorated_function: called function (original)
		:param args: args with which function is called
		:param kwargs: kwargs with which function is called

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_value(decorated_function=lambda x: callable(x))
	def get_cache(self, decorated_function, *args, **kwargs):
		""" Get cache entry (:class:`.WCacheStorage.CacheEntry`) for the specified arguments

		:param decorated_function: called function (original)
		:param args: args with which function is called
		:param kwargs: kwargs with which function is called

		:return: WCacheStorage.CacheEntry
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_value(decorated_function=lambda x: x is None or callable(x))
	def clear(self, decorated_function=None):
		""" Remove results from this storage

		:param decorated_function: if specified, then results will be removed for this function only

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@verify_value('paranoid', decorated_function=lambda x: callable(x))
	def has(self, decorated_function, *args, **kwargs):
		""" Check if there is a result for given function

		:param decorated_function: called function (original)
		:param args: args with which function is called
		:param kwargs: kwargs with which function is called

		:return: None
		"""
		return self.get_cache(decorated_function, *args, **kwargs).has_value

	@verify_value('paranoid', decorated_function=lambda x: callable(x))
	def get_result(self, decorated_function, *args, **kwargs):
		""" Get result from storage for specified function. Will raise an exception
		(:class:`.WCacheStorage.CacheMissedException`) if there is no cached result.

		:param decorated_function: called function (original)
		:param args: args with which function is called
		:param kwargs: kwargs with which function is called

		:return: (any type, even None)
		"""
		cache_entry = self.get_cache(decorated_function, *args, **kwargs)
		if cache_entry.has_value is False:
			raise WCacheStorage.CacheMissedException('No cache record found')
		return cache_entry.cached_value


class WGlobalSingletonCacheStorage(WCacheStorage):
	""" Simple storage that acts as global singleton. Result (singleton) is saved on the very first call. It doesn't
	matter with which parameters function was called, result will be the same for all the rest calls.
	"""

	def __init__(self):
		""" Construct new storage
		"""
		self._storage = {}

	@verify_value(decorated_function=lambda x: callable(x))
	def put(self, result, decorated_function, *args, **kwargs):
		""" :meth:`WCacheStorage.put` method implementation
		"""
		self._storage[decorated_function] = result

	@verify_value(decorated_function=lambda x: callable(x))
	def has(self, decorated_function, *args, **kwargs):
		""" :meth:`WCacheStorage.has` method implementation
		"""
		return decorated_function in self._storage.keys()

	@verify_value(decorated_function=lambda x: callable(x))
	def get_result(self, decorated_function, *args, **kwargs):
		""" :meth:`WCacheStorage.get_result` method implementation
		"""
		try:
			return self._storage[decorated_function]
		except KeyError:
			raise WCacheStorage.CacheMissedException('No cache record found')

	@verify_value('paranoid', decorated_function=lambda x: callable(x))
	def get_cache(self, decorated_function, *args, **kwargs):
		""" :meth:`WCacheStorage.get_cache` method implementation
		"""
		has_value = self.has(decorated_function, *args, **kwargs)
		cached_value = None
		if has_value is True:
			cached_value = self.get_result(decorated_function, *args, **kwargs)
		return WCacheStorage.CacheEntry(has_value=has_value, cached_value=cached_value)

	@verify_value(decorated_function=lambda x: x is None or callable(x))
	def clear(self, decorated_function=None):
		""" :meth:`WCacheStorage.clear` method implementation
		"""
		if decorated_function is not None and decorated_function in self._storage:
			self._storage.pop(decorated_function)
		else:
			self._storage.clear()


class WInstanceSingletonCacheStorage(WCacheStorage):
	""" This storage acts similar to :class:`.WGlobalSingletonCacheStorage` storage, but works with bounded
	methods only (class methods or object method). For every object it keeps results with "cache-record" class
	(:class:`.WInstanceSingletonCacheStorage.InstanceCacheRecord`), this class (is used by default) saves
	the very first result and returns it every time. For example, by default if we have two object derived
	from the same class, and the same method is called, then this storage will keep two separate results,
	one for each instance.

	Exact behaviour can be tweaked through :class:`.WInstanceSingletonCacheStorage.InstanceCacheRecord` inheritance.
	So derived "cache-records" classes can do things in there own way, they may save every called result, or may not
	save anything.

	This class was extended to support internal statistics with cache hits and misses. Still, this class is not
	thread safe, but accessing statistics from a separate thread should work. Statistics is calculated for
	records that was fetch through :meth:`WInstanceSingletonCacheStorage.get_cache` method only

	:note: This implementation uses weakrefs, so memory leak doesn't happen (here).
	"""

	class InstanceCacheRecord:
		""" Class is used to save cached results for the specified method and for the single instance. This
		class saves the very first result only. This class uses :class:`.WCacheStorage.CacheEntry` the same way
		as :class:`.WCacheStorage` storage does - it help to determine, whether there is a cached value or not.

		Because derived class constructor signature may differ from this class constructor signature, then
		in order to create cache record there should be a unified method, which is
		:meth:`.WInstanceSingletonCacheStorage.InstanceCacheRecord.create`
		"""

		@verify_value('paranoid', decorated_function=lambda x: callable(x))
		def __init__(self, result, decorated_function):
			""" Create new cache record

			:param result: result to keep
			:param decorated_function: called bounded method (original)
			"""
			self.__decorated_function = decorated_function
			self.__result = result

		def decorated_function(self):
			""" Return original method

			:return: bounded method
			"""
			return self.__decorated_function

		def cache_entry(self, *args, **kwargs):
			""" Return cache entry for the specified arguments

			:param args: args with which bounded method was called
			:param kwargs: kwargs with which bounded method was called

			:return: WCacheStorage.CacheEntry
			"""
			return WCacheStorage.CacheEntry(has_value=True, cached_value=self.__result)

		def update(self, result, *args, **kwargs):
			""" Update (or add other one) result, that was generated with specified arguments

			:param result: result to keep
			:param args: args with which bounded method was called
			:param kwargs: kwargs with which bounded method was called

			:return: None
			"""
			self.__result = result

		@classmethod
		@verify_value('paranoid', decorated_function=lambda x: callable(x))
		def create(cls, result, decorated_function, *args, **kwargs):
			""" Create new "cache-record" for the specified arguments

			:param result: result to keep
			:param decorated_function: called bounded method
			:param args: args with which bounded method was called
			:param kwargs: kwargs with which bounded method was called

			:return: WInstanceSingletonCacheStorage.InstanceCacheRecord
			"""
			return cls(result, decorated_function)

	@verify_type(statistic=bool)
	def __init__(self, cache_record_cls=None, statistic=False):
		""" Construct new storage

		:param cache_record_cls: class for keeping cache
		:param statistic: whether to store statistics about cache hits and misses or not
		"""
		self._storage = {}
		self._cache_record_cls = None
		if cache_record_cls is not None:
			if issubclass(cache_record_cls, WInstanceSingletonCacheStorage.InstanceCacheRecord) is False:
				raise TypeError('Invalid cache record class')
			self._cache_record_cls = cache_record_cls
		else:
			self._cache_record_cls = WInstanceSingletonCacheStorage.InstanceCacheRecord
		self.__statistic = statistic
		self.__cache_missed = 0 if self.__statistic is True else None
		self.__cache_hit = 0 if self.__statistic is True else None

	@verify_value('paranoid', decorated_function=lambda x: callable(x))
	def __check(self, decorated_function, *args, **kwargs):
		""" Check whether function is a bounded method or not. If check fails then exception is raised

		:param decorated_function: called function (original)
		:param args: args with which function is called
		:param kwargs: kwargs with which function is called
		:return: None
		"""
		# TODO replace this function with decorator which can be turned off like verify_* does
		if len(args) >= 1:
			obj = args[0]
			function_name = decorated_function.__name__
			if hasattr(obj, function_name) is True:
				fn = getattr(obj, function_name)
				if callable(fn) and fn.__self__ == obj:
					return

		raise RuntimeError('Only bounded methods are allowed')

	@verify_value(decorated_function=lambda x: callable(x))
	def put(self, result, decorated_function, *args, **kwargs):
		""" :meth:`WCacheStorage.put` method implementation
		"""
		self.__check(decorated_function, *args, **kwargs)

		ref = weakref.ref(args[0])
		if decorated_function not in self._storage:
			cache_entry = self._cache_record_cls.create(result, decorated_function, *args, **kwargs)
			self._storage[decorated_function] = [{'instance': ref, 'result': cache_entry}]
		else:
			instance_found = False
			for i in self._storage[decorated_function]:
				if i['instance']() == args[0]:
					cache_entry = i['result']
					cache_entry.update(result, *args, **kwargs)
					instance_found = True
					break
			if instance_found is False:
				cache_entry = self._cache_record_cls.create(result, decorated_function, *args, **kwargs)
				self._storage[decorated_function].append({'instance': ref, 'result': cache_entry})

		def finalize_ref():
			if decorated_function in self._storage:
				fn_list = self._storage[decorated_function]
				if len(fn_list) == 1 and fn_list[0]['instance'] == ref:
					del self._storage[decorated_function]

				for i in range(len(fn_list)):
					if fn_list[i]['instance'] == ref:
						fn_list.pop(i)
						return

		weakref.finalize(args[0], finalize_ref)

	@verify_value(decorated_function=lambda x: callable(x))
	def get_cache(self, decorated_function, *args, **kwargs):
		""" :meth:`WCacheStorage.get_cache` method implementation
		"""
		self.__check(decorated_function, *args, **kwargs)
		if decorated_function in self._storage:
			for i in self._storage[decorated_function]:
				if i['instance']() == args[0]:
					result = i['result'].cache_entry(*args, **kwargs)
					if self.__statistic is True:
						if result.has_value is True:
							self.__cache_hit += 1
						else:
							self.__cache_missed += 1
					return result

		if self.__statistic is True:
			self.__cache_missed += 1

		return WCacheStorage.CacheEntry()

	@verify_value(decorated_function=lambda x: x is None or callable(x))
	def clear(self, decorated_function=None):
		""" :meth:`WCacheStorage.clear` method implementation (Clears statistics also)
		"""
		if decorated_function is not None and decorated_function in self._storage:
			self._storage.pop(decorated_function)
		else:
			self._storage.clear()

		if self.__statistic is True:
			self.__cache_missed = 0
			self.__cache_hit = 0

	def cache_missed(self):
		""" Return cache misses (return None if class was constructed without 'statistic' flag)

		:return: int or None
		"""
		return self.__cache_missed

	def cache_hit(self):
		""" Return cache hits (return None if class was constructed without 'statistic' flag)

		:return: int or None
		"""
		return self.__cache_hit


@verify_type(storage=(None, WCacheStorage))
@verify_value(validator=lambda x: x is None or callable(x))
def cache_control(validator=None, storage=None):
	""" Decorator that is used for caching result.

	:param validator: function, that has following signature (decorated_function, \*args, \*\*kwargs), where \
	decorated_function - original function, args - function arguments, kwargs - function keyword arguments. \
	This function must return True if cache is valid (old result must be use if it there is one), or False - to \
	generate and to store new result. So function that always return True can be used as singleton. And function \
	that always return False won't cache anything at all. By default (if no validator is specified), it presumes \
	that cache is always valid.
	:param storage: storage that is used for caching results. see :class:`.WCacheStorage` class.

	:return: decorated function
	"""

	def default_validator(*args, **kwargs):
		return True

	if validator is None:
		validator = default_validator

	if storage is None:
		storage = WGlobalSingletonCacheStorage()

	def first_level_decorator(decorated_function):
		def second_level_decorator(original_function, *args, **kwargs):

			validator_check = validator(original_function, *args, **kwargs)
			cache_entry = storage.get_cache(original_function, *args, **kwargs)

			if validator_check is not True or cache_entry.has_value is False:
				result = original_function(*args, **kwargs)
				storage.put(result, original_function, *args, **kwargs)
				return result
			else:
				return cache_entry.cached_value

		return decorator(second_level_decorator)(decorated_function)
	return first_level_decorator



# -*- coding: utf-8 -*-
# wasp_general/thread.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from threading import Lock
from decorator import decorator

from wasp_general.verify import verify_type, verify_value


class WCriticalSectionError(Exception):
	pass


@verify_type(blocking=bool, timeout=(int, float, None), raise_exception=bool)
@verify_value(lock_fn=lambda x: callable(x))
@verify_value(timeout=lambda x: x is None or x > 0)
def critical_section_dynamic_lock(lock_fn=None, blocking=True, timeout=None, raise_exception=True):

	if blocking is False or timeout is None:
		timeout = -1

	def first_level_decorator(decorated_function):
		def second_level_decorator(original_function, *args, **kwargs):
			lock = lock_fn(*args, **kwargs)
			if lock.acquire(blocking=blocking, timeout=timeout) is True:
				try:
					result = original_function(*args, **kwargs)
					return result
				finally:
					lock.release()
			elif raise_exception is True:
				WCriticalSectionError('Unable to lock critical section\n')

		return decorator(second_level_decorator)(decorated_function)
	return first_level_decorator


@verify_type('paranoid', lock=Lock().__class__, blocking=bool, timeout=(int, float, None), raise_exception=bool)
@verify_value('paranoid', timeout=lambda x: x is None or x > 0)
@verify_type(lock=Lock().__class__)
def critical_section_lock(lock=None, blocking=True, timeout=None, raise_exception=True):

	def lock_getter(*args, **kwargs):
		return lock

	return critical_section_dynamic_lock(
		lock_fn=lock_getter, blocking=blocking, timeout=timeout, raise_exception=raise_exception
	)


class WCriticalResource:

	def __init__(self):
		self.__lock = Lock()

	def thread_lock(self):
		return self.__lock

	@staticmethod
	@verify_type('paranoid', blocking=bool, timeout=(int, float, None), raise_exception=bool)
	@verify_value('paranoid', timeout=lambda x: x is None or x > 0)
	def critical_section(blocking=True, timeout=None, raise_exception=True):

		def lock_getter(self, *args, **kwargs):
			if isinstance(self, WCriticalResource) is False:
				raise TypeError(
					'Invalid object type. It must be inherited from WCriticalResource class and'
					' decorated method must be bounded'
				)
			return self.thread_lock()

		return critical_section_dynamic_lock(
			lock_fn=lock_getter, blocking=blocking, timeout=timeout, raise_exception=raise_exception
		)



# -*- coding: utf-8 -*-
# wasp_general/network/clients/base.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from enum import Enum
from abc import abstractmethod, abstractclassmethod

from wasp_general.network.clients.proto import WNetworkClientCapabilityProto, WNetworkClientProto
from wasp_general.verify import verify_type
from wasp_general.uri import WURI


class WCommonNetworkClientCapability(Enum):
	current_dir = 'current_dir'
	change_dir = 'change_dir'
	list_dir = 'list_dir'
	make_dir = 'make_dir'
	upload_file = 'upload_file'
	remove_file = 'remove_file'


# noinspection PyAbstractClass
class WBasicNetworkClientCapability(WNetworkClientCapabilityProto):

	def __init__(self, network_agent):
		WNetworkClientCapabilityProto.__init__(self)
		self.__network_agent = network_agent

	def network_agent(self):
		return self.__network_agent

	def capability_id(self):
		return self.common_capability().value

	@abstractclassmethod
	def common_capability(cls):
		raise NotImplementedError('This method is abstract')

	@classmethod
	def create_capability(cls, network_agent):
		return cls(network_agent)


# noinspection PyAbstractClass
class WBasicNetworkClientProto(WNetworkClientProto):

	def __init__(self, uri):
		self.__uri = uri
		self.__capabilities = \
			{x.common_capability(): x.create_capability(self) for x in self.agent_capabilities()}
		self.__is_closed = False

	def uri(self):
		return self.__uri

	def capabilities(self):
		return self.__capabilities.values()

	def request(self, capability_id, *args, **kwargs):
		if self.is_closed() is True:
			raise RuntimeError('Operation requested on this closed client')

		if self.is_capable(capability_id) is True:
			common_capability = WCommonNetworkClientCapability(capability_id)
			for capability in self.capabilities():
				if capability.common_capability() == common_capability:
					return capability.request(*args, **kwargs)
		raise RuntimeError('Unable to execute unsupported capability: %s' % str(capability_id))

	def is_closed(self):
		return self.__is_closed

	def close(self):
		if self.__is_closed is False:
			self._close()
		self.__is_closed = True

	@classmethod
	def is_capable(cls, capability_id):
		common_capability = WCommonNetworkClientCapability(capability_id)
		for agent_cap in cls.agent_capabilities():
			if agent_cap.common_capability() == common_capability:
				return True
		return False

	@classmethod
	@verify_type(uri=WURI)
	def create_handler(cls, uri):
		return cls(uri)

	@abstractmethod
	def _close(self):
		raise NotImplementedError('This method is abstract')

	@abstractclassmethod
	def agent_capabilities(cls):
		raise NotImplementedError('This method is abstract')


# noinspection PyAbstractClass
class WBasicNetworkClientChangeDirCapability(WBasicNetworkClientCapability):

	@classmethod
	def common_capability(cls):
		return WCommonNetworkClientCapability.change_dir


# noinspection PyAbstractClass
class WBasicNetworkClientListDirCapability(WBasicNetworkClientCapability):

	@classmethod
	def common_capability(cls):
		return WCommonNetworkClientCapability.list_dir


# noinspection PyAbstractClass
class WBasicNetworkClientMakeDirCapability(WBasicNetworkClientCapability):

	@classmethod
	def common_capability(cls):
		return WCommonNetworkClientCapability.make_dir


# noinspection PyAbstractClass
class WBasicNetworkClientCurrentDirCapability(WBasicNetworkClientCapability):

	@classmethod
	def common_capability(cls):
		return WCommonNetworkClientCapability.current_dir


# noinspection PyAbstractClass
class WBasicNetworkClientUploadFileCapability(WBasicNetworkClientCapability):

	@classmethod
	def common_capability(cls):
		return WCommonNetworkClientCapability.upload_file


# noinspection PyAbstractClass
class WBasicNetworkClientRemoveFileCapability(WBasicNetworkClientCapability):

	@classmethod
	def common_capability(cls):
		return WCommonNetworkClientCapability.remove_file
